"""Autograding process internals for Otter-Grader"""

import os
import json
import pandas as pd
import pickle

from .runners import create_runner
from .utils import OtterRuntimeError
from ...version import LOGO_WITH_VERSION
from ...utils import chdir


def main(autograder_dir, **kwargs):
    """
    Run the autograding process.

    Args:
        autograder_dir (``str``): the absolute path of the directory in which autograding is occurring
            (e.g. on Gradescope, this is ``/autograder``)
        **kwargs: keyword arguments for updating configurations in the default configurations 
            ``otter.run.run_autograder.constants.DEFAULT_OPTIONS``; these values override anything
            present in ``otter_config.json``
    """
    config_fp = os.path.join(autograder_dir, "source", "otter_config.json")
    if os.path.isfile(config_fp):
        with open(config_fp, encoding="utf-8") as f:
            config = json.load(f)
    else:
        config = {}

    config["autograder_dir"] = autograder_dir

    runner = create_runner(config, **kwargs)

    if runner.get_option("logo"):
        # ASCII 8207 is an invisible non-whitespace character; this should prevent gradescope from
        # incorrectly left-stripping the whitespace at the beginning of the logo
        print(f"{chr(8207)}\n", LOGO_WITH_VERSION, "\n", sep="")

    abs_ag_path = os.path.abspath(runner.get_option("autograder_dir"))
    with chdir(abs_ag_path):        
        try:
            runner.prepare_files()
            scores = runner.run()
            with open("results/results.pkl", "wb+") as f:
                    pickle.dump(scores, f)

            output = scores.to_gradescope_dict(runner.get_options())

        except OtterRuntimeError as e:
            output = {
                "score": 0,
                "stdout_visibility": "hidden",
                "tests": [
                    {
                        "name": "Autograder Error",
                        "output": f"Otter encountered an error when grading this submission:\n\n{e}",
                    },
                ],
            }
            raise e

        finally:
            if "output" in vars():
                with open("./results/results.json", "w+") as f:
                    json.dump(output, f, indent=4)                

    print("\n\n", end="")

    df = pd.DataFrame(output["tests"])

    if runner.get_option("print_score"):
        total, possible = df["score"].sum(), df["max_score"].sum()
        if "score" in output:
            total, possible = output["score"], runner.get_option("points_possible") or possible
        perc = total / possible * 100
        print(f"Total Score: {total:.3f} / {possible:.3f} ({perc:.3f}%)\n")

    if runner.get_option("print_summary"):
        pd.set_option("display.max_rows", None)  # print all rows of the dataframe
        if "output" in df.columns:
            df.drop(columns=["output"], inplace=True)
        if "visibility" in df.columns:
            df.drop(columns=["visibility"], inplace=True)

        print(df)



"""Solution removal for Otter Assign"""

import copy
import nbformat
import re

from .r_adapter import solutions as r_solutions
from .utils import get_notebook_language, get_source, has_tag, is_cell_type, remove_output, \
    remove_tag


SOLUTION_CELL_TAG = "otter_assign_solution_cell"


def has_seed(cell):
    """
    Determine whether a cell contains a seed line (a line ending in ``# SEED``).

    Args:
        cell (``nbformat.NotebookNode``): the cell

    Returns:
        ``bool``: whether the cell has a seed line
    """
    if not is_cell_type(cell, "code"):
        return False
    source = get_source(cell)
    return source and any([l.strip().endswith('# SEED') for l in source])


def overwrite_seed_vars(nb, seed_variable, seed):
    """
    Overwrite any assignments of the variable named ``seed_variable`` with the value ``seed`` in a
    notebook.

    Args:
        nb (``nbformat.NotebookNode``): the notebook to edit
        seed_variable (``str``): the variable to look for
        seed (``int``): the value to set for ``seed_variable``

    Returns:
        ``nbformat.NotebookNode``: the notebook with the variable substitutions made
    """
    nb = copy.deepcopy(nb)
    for cell in nb["cells"]:
        source = get_source(cell)
        for i, line in enumerate(source):
            match = re.match(fr"(\s*){seed_variable}\s*(=|<-)\s*", line)
            if  match:
                source[i] = match.group(1) + f"{seed_variable} {match.group(2)} {seed}"
        cell["source"] = "\n".join(source)
    return nb


solution_assignment_regex = re.compile(r"(\s*[a-zA-Z0-9_. ]*=).* ?# ?SOLUTION")
def solution_assignment_sub(match):
    """
    Substitutes the first matching group  with ` ...`
    """
    prefix = match.group(1)
    return prefix + ' ...'


solution_line_regex = re.compile(r"(\s*).* ?# ?SOLUTION")
def solution_line_sub(match):
    """
    Substitutes the first matching group  with `...`
    """
    prefix = match.group(1)
    return prefix + '...'


begin_solution_regex = re.compile(r"(\s*)# BEGIN SOLUTION( NO PROMPT)?")
skip_suffixes = ['# SOLUTION NO PROMPT', '# BEGIN PROMPT', '# END PROMPT', '# SEED']

SUBSTITUTIONS = {
    "python":  [
        (solution_assignment_regex, solution_assignment_sub),
        (solution_line_regex, solution_line_sub),
    ],
    "r":  r_solutions.SUBSTITUTIONS,
}


# TODO: comments, docstrings
def replace_solutions(lines, lang):
    """
    Replaces solutions in ``lines``
    
    Args:
        lines (``list`` of ``str``): solutions as a list of strings

    Returns:
        ``list`` of ``str``: stripped version of lines without solutions
    """
    stripped = []
    solution = False
    for line in lines:

        # ...
        if any(line.rstrip().endswith(s) for s in skip_suffixes):
            continue

        # ...
        if solution and not line.rstrip().endswith('# END SOLUTION'):
            continue

        # ...
        if line.rstrip().endswith('# END SOLUTION'):
            assert solution, f"END SOLUTION without BEGIN SOLUTION in {lines}"
            solution = False
            continue

        # ...
        begin_solution = begin_solution_regex.match(line)
        if begin_solution:
            assert not solution, f"Nested BEGIN SOLUTION in {lines}"
            solution = True
            if not begin_solution.group(2):
                line = begin_solution.group(1) + '...'
            else:
                continue
        for exp, sub in SUBSTITUTIONS[lang]:
            m = exp.match(line)
            if m:
                line = sub(m)
                break
        
        stripped.append(line)
    
    assert not solution, f"BEGIN SOLUTION without END SOLUTION in {lines}"
    
    return stripped


def remove_ignored_lines(lines):
    """
    Removes ignored lines in ``lines``
    
    Args:
        lines (``list`` of ``str``): cell source as a list of strings

    Returns:
        ``list`` of ``str``: stripped version of lines without ignored lines
    """
    ignore_suffix = "# IGNORE"
    stripped = []
    in_block = False
    for line in lines:

        # ...
        if line.rstrip().endswith(ignore_suffix):
            continue

        # ...
        if in_block and not line.rstrip().endswith('# END IGNORE'):
            continue

        # ...
        if line.rstrip().endswith('# END IGNORE'):
            assert in_block, f"END IGNORE without BEGIN IGNORE in {lines}"
            in_block = False
            continue

        # ...
        if re.match(r"\s*#\s*BEGIN\s*IGNORE\s*", line, flags=re.IGNORECASE):
            assert not in_block, f"Nested BEGIN IGNORE in {lines}"
            in_block = True
            continue

        stripped.append(line)
    
    assert not in_block, f"BEGIN IGNORE without END IGNORE in {lines}"
    
    return stripped


def strip_ignored_lines(nb):
    """
    Write a notebook with ignored lines stripped
    
    Args:
        nb (``nbformat.NotebookNode``): the notebook to have ignored lines stripped
    """
    for i, cell in enumerate(nb['cells']):
        cell['source'] = '\n'.join(remove_ignored_lines(get_source(cell)))
    return nb


def strip_solutions_and_output(nb):
    """
    Write a notebook with solutions stripped and outputs cleared
    
    Args:
        nb (``nbformat.NotebookNode``): the notebook to have solutions stripped
    """
    md_solutions = []
    lang = get_notebook_language(nb)
    for i, cell in enumerate(nb['cells']):
        if has_tag(cell, SOLUTION_CELL_TAG):
            if is_cell_type(cell, "code"):
                cell['source'] = '\n'.join(replace_solutions(get_source(cell), lang))
            elif is_cell_type(cell, "markdown"):
                md_solutions.append(i)
            nb['cells'][i] = remove_tag(cell, SOLUTION_CELL_TAG)

    md_solutions.reverse()
    for i in md_solutions:
        del nb['cells'][i]
    
    # remove output from student version
    remove_output(nb)
    
    return nb



"""
Question configurations for Otter Assign
"""

import copy
import yaml

from .constants import BLOCK_QUOTE, ALLOWED_NAME
from .utils import get_source, lock, get_spec, EmptyCellException
from ..utils import convert_config_description_dict


_DEFAULT_QUESTION_CONFIGURATIONS_WITH_DESCRIPTIONS = [
    {
        "key": "name",
        "description": "(required) the path to a requirements.txt file",
        "required": True,
    },
    {
        "key": "manual",
        "description": "whether this is a manually-graded question",
        "default": False,
    },
    {
        "key": "points",
        "description": "how many points this question is worth; defaults to 1 internally",
        "default": None,
    },
    {
        "key": "check_cell",
        "description": "whether to include a check cell after this question (for autograded questions only)",
        "default": True,
    },
    {
        "key": "export",
        "description": "whether to force-include this question in the exported PDF",
        "default": False,
    },
]

DEFAULT_QUESTION_CONFIGURATIONS = convert_config_description_dict(_DEFAULT_QUESTION_CONFIGURATIONS_WITH_DESCRIPTIONS)


def create_question_config(parsed_config):
    """
    """
    config = DEFAULT_QUESTION_CONFIGURATIONS.copy()
    config.update(parsed_config)
    if "name" not in config:
        raise ValueError(f"Question name not specified in config YAML: {parsed_config}")
    return config

def is_question_cell(cell):
    """
    Returns whether cell contains BEGIN QUESTION in a block quote
    
    Args:
        cell (``nbformat.NotebookNode``): a notebook cell
    
    Returns:
        ``bool``: whether the current cell is a question definition cell
    """
    if cell.cell_type != 'markdown':
        return False
    return get_spec(get_source(cell), "question") is not None


def read_question_metadata(cell):
    """
    Returns parsed question metadata from a question cell
    
    Args:
        cell (``nbformat.NotebookNode``): the question cell
    
    Returns:
        ``dict``: question metadata
    """
    source = get_source(cell)
    begin_question_line = get_spec(source, "question")
    i, lines = begin_question_line + 1, []
    while source[i].strip() != BLOCK_QUOTE:
        lines.append(source[i])
        i = i + 1
    metadata = yaml.full_load('\n'.join(lines))
    assert ALLOWED_NAME.match(metadata.get('name', '')), metadata
    return metadata



"""Otter Check command-line utility"""

import os

from glob import glob
from jinja2 import Template

from .logs import LogEntry, EventType
from .notebook import _OTTER_LOG_FILENAME

from ..execute import grade_notebook, check
from ..utils import block_print


def _log_event(event_type, results=[], question=None, success=True, error=None):
	"""
	Logs an event

	Args:
		event_type (``otter.logs.EventType``): the type of event
		results (``list`` of ``otter.test_files.abstract_test.TestCollectionResults``, optional): the 
			results of any checks recorded by the entry
		question (``str``, optional): the question name for this check
		success (``bool``, optional): whether the operation was successful
		error (``Exception``, optional): the exception thrown by the operation, if applicable
	"""
	LogEntry(
		event_type,
		results=results,
		question=question, 
		success=success, 
		error=error
	).flush_to_file(_OTTER_LOG_FILENAME)


def main(file,  *, tests_path="./tests", question=None, seed=None):
	"""
	Runs Otter Check

	Args:
		file (``str``): path to the file to check
		tests_path (``str``): path to tests directory
		question (``str``): test name to run; ``None`` if all tests should be run
		seed (``int``): a seed to set before execution
		**kwargs: ignored kwargs (a remnant of how the argument parser is built)
	"""
	try:
		if question:
			test_path = os.path.join(tests_path, question + ".py")
			assert os.path.isfile(test_path), "Test {} does not exist".format(question)
			qs = [test_path]
		else:
			qs = glob(os.path.join(tests_path, "*.py"))

		assert os.path.isfile(file), "{} is not a file".format(file)
		assert file[-6:] == ".ipynb" or file[-3:] == ".py", "{} is not a Jupyter Notebook or Python file".format(file)

		script = file[-3:] == ".py"

		with block_print():
			results = grade_notebook(
				file,
				tests_glob=qs,
				script=script,
				seed=seed
			)

		if results.total / results.possible == 1:
			output = "All tests passed!"
		else:
			output = "\n".join(repr(test_file) for _, test_file in results.results.items())

		print(output)

	except Exception as e:
		_log_event(EventType.CHECK, success=False, error=e)
		raise e
			
	else:
		_log_event(EventType.CHECK, results=results)



###############################
##### Tests for otter run #####
###############################

import os
import unittest
import subprocess
import json
import shutil
import nbformat
import nbconvert

from subprocess import PIPE
from glob import glob
from unittest import mock
from shutil import copyfile

from otter.run.run_autograder import main as run_autograder

from . import TestCase


NBFORMAT_VERSION = 4
TEST_FILES_PATH = "test/test-run/"


class TestRun(TestCase):

    def setUp(self):
        super().setUp()

        self.cwd = os.getcwd()

        with open(TEST_FILES_PATH + "autograder/source/otter_config.json") as f:
            self.config = json.load(f)
        
        self.expected_results = {
            "tests": [
                {
                    "name": "Public Tests",
                    "visibility": "visible",
                    "output": "q1 results: All test cases passed!\n\nq2 results:\n    q2 - 1 result:\n        Trying:\n            negate(True)\n        Expecting:\n            False\n        **********************************************************************\n        Line 2, in q2 0\n        Failed example:\n            negate(True)\n        Expected:\n            False\n        Got:\n            True\n\n    q2 - 2 result:\n        Trying:\n            negate(False)\n        Expecting:\n            True\n        **********************************************************************\n        Line 2, in q2 1\n        Failed example:\n            negate(False)\n        Expected:\n            True\n        Got:\n            False\n\nq3 results: All test cases passed!\n\nq4 results: All test cases passed!\n\nq6 results: All test cases passed!\n\nq7 results: All test cases passed!"
                },
                {
                    "name": "q1",
                    "score": 0.0,
                    "max_score": 0.0,
                    "visibility": "hidden",
                    "output": "q1 results: All test cases passed!"
                },
                {
                    "name": "q2",
                    "score": 0,
                    "max_score": 2.0,
                    "visibility": "hidden",
                    "output": "q2 results:\n    q2 - 1 result:\n        Trying:\n            negate(True)\n        Expecting:\n            False\n        **********************************************************************\n        Line 2, in q2 0\n        Failed example:\n            negate(True)\n        Expected:\n            False\n        Got:\n            True\n\n    q2 - 2 result:\n        Trying:\n            negate(False)\n        Expecting:\n            True\n        **********************************************************************\n        Line 2, in q2 1\n        Failed example:\n            negate(False)\n        Expected:\n            True\n        Got:\n            False\n\n    q2 - 3 result:\n        Trying:\n            negate(\"\")\n        Expecting:\n            True\n        **********************************************************************\n        Line 2, in q2 2\n        Failed example:\n            negate(\"\")\n        Expected:\n            True\n        Got:\n            ''\n\n    q2 - 4 result:\n        Trying:\n            negate(1)\n        Expecting:\n            False\n        **********************************************************************\n        Line 2, in q2 3\n        Failed example:\n            negate(1)\n        Expected:\n            False\n        Got:\n            1"
                },
                {
                    "name": "q3",
                    "score": 2.0,
                    "max_score": 2.0,
                    "visibility": "hidden",
                    "output": "q3 results: All test cases passed!"
                },
                {
                    "name": "q4",
                    "score": 1.0,
                    "max_score": 1.0,
                    "visibility": "hidden",
                    "output": "q4 results: All test cases passed!"
                },
                {
                    "name": "q6",
                    "score": 2.5,
                    "max_score": 5.0,
                    "visibility": "hidden",
                    "output": "q6 results:\n    q6 - 1 result:\n        Test case passed!\n\n    q6 - 2 result:\n        Trying:\n            fib = fiberator()\n        Expecting nothing\n        ok\n        Trying:\n            for _ in range(10):\n                print(next(fib))\n        Expecting:\n            0\n            1\n            1\n            2\n            3\n            5\n            8\n            13\n            21\n            34\n        **********************************************************************\n        Line 3, in q6 1\n        Failed example:\n            for _ in range(10):\n                print(next(fib))\n        Expected:\n            0\n            1\n            1\n            2\n            3\n            5\n            8\n            13\n            21\n            34\n        Got:\n            0\n            1\n            1\n            1\n            2\n            3\n            5\n            8\n            13\n            21"
                },
                {
                    "name": "q7",
                    "score": 1.0,
                    "max_score": 1.0,
                    "visibility": "hidden",
                    "output": "q7 results: All test cases passed!"
                }
            ],
            "output": "Students are allowed 1 submissions every 1 days. You have 0 submissions in that period."
        }

    def test_notebook(self):
        run_autograder(self.config['autograder_dir'])

        with open(TEST_FILES_PATH + "autograder/results/results.json") as f:
            actual_results = json.load(f)

        self.assertEqual(actual_results, self.expected_results, f"Actual results did not matched expected:\n{actual_results}")

    def test_script(self):
        nb_path = TEST_FILES_PATH + "autograder/submission/fails2and6H.ipynb"
        nb = nbformat.read(nb_path, as_version=NBFORMAT_VERSION)
        os.remove(nb_path)

        try:
            py, _ = nbconvert.export(nbconvert.PythonExporter, nb)
            
            # remove magic commands
            py = "\n".join(l for l in py.split("\n") if not l.startswith("get_ipython"))
            
            with open(TEST_FILES_PATH + "autograder/submission/fails2and6H.py", "w+") as f:
                f.write(py)

            run_autograder(self.config['autograder_dir'])

            with open(TEST_FILES_PATH + "autograder/results/results.json") as f:
                actual_results = json.load(f)

            self.assertEqual(actual_results, self.expected_results, f"Actual results did not matched expected:\n{actual_results}")

            self.deletePaths([TEST_FILES_PATH + "autograder/submission/fails2and6H.py"])

            with open(nb_path, "w+") as f:
                nbformat.write(nb, f)
        
        except:
            os.chdir(self.cwd)
            self.deletePaths([TEST_FILES_PATH + "autograder/submission/fails2and6H.py"])
            with open(nb_path, "w+") as f:
                nbformat.write(nb, f)
            
            raise

    def tearDown(self):
        os.chdir(self.cwd)
        self.deletePaths([
            TEST_FILES_PATH + "autograder/results/results.json",
            TEST_FILES_PATH + "autograder/results/results.pkl",
            TEST_FILES_PATH + "autograder/__init__.py",
            TEST_FILES_PATH + "autograder/submission/test",
            TEST_FILES_PATH + "autograder/submission/tests",
            TEST_FILES_PATH + "autograder/submission/__init__.py",
            TEST_FILES_PATH + "autograder/submission/.OTTER_LOG",
        ])



"""Transformations to apply to a submission before execution"""

import copy
import nbformat


IGNORE_CELL_TAG = "otter_ignore"
CELL_METADATA_KEY = "otter"


def script_to_notebook(script):
    """
    Convert a Python script (a string) to a notebook with ``nbformat``.

    Args:
        script (``str``): the script

    Returns:
        ``nbformat.NotebookNode``: a notebook with a single code cell containing the script
    """
    nb = nbformat.v4.new_notebook()
    nb.cells.append(nbformat.v4.new_code_cell(script))
    return nb


def filter_ignored_cells(nb):
    """
    Filter out all cells in the notebook ``nb`` that are tagged with ``otter_ignore`` or have the
    ``ignore`` key of their Otter cell metadata set to true.

    Args:
        nb (``nbformat.NotebookNode``): the notebook

    Returns:
        ``nbformat.NotebookNode``: the notebook with ignored cells removed
    """
    nb = copy.deepcopy(nb)

    to_delete = []
    for i, cell in enumerate(nb["cells"]):
        metadata = cell.get("metadata", {})
        tags = metadata.get("tags", [])

        if IGNORE_CELL_TAG in tags or metadata.get(CELL_METADATA_KEY, {}).get("ignore", False):
            to_delete.append(i)

    to_delete.reverse()
    for i in to_delete:
        del nb["cells"][i]

    return nb


def create_collected_check_cell(cell, check_results_list_name, notebook_class_name, test_dir):
    """
    Generate a string of calls to ``otter.Notebook.check`` collecting the results in a list called
    ``check_results_list_name`` based on the cell metadata.

    Note that this string is formatted with surrounding newlines, so it can be inserted into any
    Python script as-is.

    Args:
        cell (``nbformat.NotebookNode``): the code cell to which checks should be appended
        check_results_list_name (``check_results_list_name``): the name of the list to append the 
            results to
        notebook_class_name(``str``): the name of the ``otter.Notebook`` class in the environment
        test_dir (``str``): the path to the directory of tests.       

    Returns:
        ``str``: the code to run the checks; if no checks are indicated, the empty string
    """
    source = ""
    otter_config = cell.get("metadata", {}).get(CELL_METADATA_KEY, {})

    if otter_config.get("tests", []):
        tests = otter_config.get("tests", [])
        for test in tests:
            source += f"\n{check_results_list_name}.append({notebook_class_name}(" + \
                f"tests_dir='{test_dir}').check('{test}'))\n"

    return source



import re

from .utils import create_cell

from ..constants import TEST_REGEX
from ..r_adapter.tests import gen_suite
from ..utils import get_source

from ...test_files.abstract_test import TestFile


def is_test_cell(cell):
    """
    Returns whether the current cell is a test cell
    
    Args:
        cell (``nbformat.NotebookNode``): an Rmd file cell

    Returns:
        ``bool``: whether the cell is a test cell
    """
    if cell.cell_type != 'code':
        return False
    source = get_source(cell)
    return source and re.match(TEST_REGEX, source[1], flags=re.IGNORECASE)


def gen_test_cell(question, tests, tests_dict, assignment):
    """
    Parses a list of test named tuples and creates a single test file. Adds this test file as a value
    to ``tests_dict`` with a key corresponding to the test's name, taken from ``question``. Returns
    a code cell that runs the check on this test.
    
    Args:
        question (``dict``): question metadata
        tests (``list`` of ``Test``): tests to be written
        tests_dict (``dict``): the tests for this assignment
        assignment (``otter.assign.assignment.Assignment``): the assignment configurations

    Returns:
        ``nbformat.NotebookNode``: code cell calling ``ottr::check`` on this test
    """
    source = f'```{{r}}\n. = ottr::check("tests/{question["name"]}.R")\n```'
    cell = create_cell("code", source)

    points = question.get('points', None)
    if isinstance(points, dict):
        points = points.get('each', 1) * len(tests)
    elif isinstance(points, list):
        if len(points) != len(tests):
            raise ValueError(
                f"Error in question {question['name']}: length of 'points' is {len(points)} but there "
                f"are {len(tests)} tests")

    # check for errors in resolving points
    tests = TestFile.resolve_test_file_points(points, tests)

    test = gen_suite(question['name'], tests, points)

    tests_dict[question['name']] = test
    return cell



"""
Support for notebook metadata test files
"""

import os
import io
import json
import doctest
import warnings
import pathlib

from contextlib import redirect_stderr, redirect_stdout
from textwrap import dedent

from .abstract_test import TestCase
from .ok_test import OKTestFile


NOTEBOOK_METADATA_KEY = "otter"


class NotebookMetadataOKTestFile(OKTestFile):
    """
    A single notebook metadata test file for Otter.

    Tests are defined in the metadata of a jupyter notebook as a JSON object with the ``otter`` key.
    The tests themselves are OK-formatted.

    .. code-block:: json

    {
        "metadata": {
            "otter": {
                "tests": {
                    "q1": {}
                }
            }
        }
    }

    Args:
        name (``str``): the name of test file
        path (``str``): the path to the test file
        test_cases (``list`` of ``TestCase``): a list of parsed tests to be run
        value (``int``, optional): the point value of this test, defaults to 1
        all_or_nothing (``bool``, optional): whether the test should be graded all-or-nothing across
            cases

    Attributes:
        name (``str``): the name of test file
        path (``str``): the path to the test file
        test_cases (``list`` of ``TestCase``): a list of parsed tests to be run
        value (``int``): the point value of this test, defaults to 1
        all_or_nothing (``bool``): whether the test should be graded all-or-nothing across
            cases
        passed_all (``bool``): whether all of the test cases were passed
        test_case_results (``list`` of ``TestCaseResult``): a list of results for the test cases in
            ``test_cases``
        grade (``float``): the percentage of ``points`` earned for this test file as a decimal
    """

    @classmethod
    def from_file(cls, path, test_name):
        """
        Parse an ok test file & return an ``OKTest``

        Args:
            path (``str``): path to ok test file

        Returns:
            ``otter.ok_parser.OKTest``: new ``OKTest`` object created from the given file
        """
        with open(path, encoding="utf-8") as f:
            nb = json.load(f)
        
        test_spec = nb["metadata"][NOTEBOOK_METADATA_KEY]["tests"]
        if test_name not in test_spec:
            raise ValueError(f"Test {test_name} not found")
        
        test_spec = test_spec[test_name]

        return cls.from_spec(test_spec, path=path)



"""
"""

import os

def zip_folder(zf, path, prefix=""):
    """
    """
    if not os.path.isabs(path):
        raise ValueError("'path' must be absolute path")
    parent_basename = os.path.basename(path)
    for file in os.listdir(path):
        child_path = os.path.join(path, file)
        if os.path.isfile(child_path):
            arcname = os.path.join(prefix, parent_basename, file)
            zf.write(child_path, arcname=arcname)
        elif os.path.isdir(child_path):
            zip_folder(zf, child_path, prefix=os.path.join(prefix, parent_basename))



"""A programmatic API for using Otter-Grader"""

__all__ = ["export_notebook", "grade_submission"]

import os
import sys
import shutil
import tempfile

from contextlib import redirect_stdout

try:
    from contextlib import nullcontext
except ImportError:
    from .utils import nullcontext  # nullcontext is new in Python 3.7

from .export import export_notebook
from .run import main as run_grader


def grade_submission(submission_path, ag_path="autograder.zip", quiet=False, debug=False):
    """
    Runs non-containerized grading on a single submission at ``submission_path`` using the autograder 
    configuration file at ``ag_path``. 
    
    Creates a temporary grading directory using the ``tempfile`` library and grades the submission 
    by replicating the autograder tree structure in that folder and running the autograder there. Does
    not run environment setup files (e.g. ``setup.sh``) or install requirements, so any requirements 
    should be available in the environment being used for grading. 
    
    Print statements executed during grading can be suppressed with ``quiet``.

    Args:
        submission_path (``str``): path to submission file
        ag_path (``str``): path to autograder zip file
        quiet (``bool``, optional): whether to suppress print statements during grading; default 
            ``False``
        debug (``bool``, optional): whether to run the submission in debug mode (without ignoring
            errors)

    Returns:
        ``otter.test_files.GradingResults``: the results object produced during the grading of the
            submission.
    """

    dp = tempfile.mkdtemp()

    if quiet:
        f = open(os.devnull, "w")
        cm = redirect_stdout(f)
    else:
        cm = nullcontext()

    # TODO: is the output_dir argument of run_grader necessary here?
    with cm:
        results = run_grader(
            submission_path, autograder=ag_path, output_dir=dp, no_logo=True, debug=debug)

    if quiet:
        f.close()

    shutil.rmtree(dp)

    return results



"""Utilities for Otter Grade"""

from python_on_whales import docker
import os
import pandas as pd
import re

from hashlib import md5


OTTER_DOCKER_IMAGE_TAG = "otter-grade"


def list_files(path):
    """
    Returns a list of all non-hidden files in a directory
    
    Args:
        path (``str``): path to a directory
    
    Returns:
        ``list`` of ``str``: list of filenames in the given directory

    """
    return [file for file in os.listdir(path) if os.path.isfile(os.path.join(path, file)) and file[0] != "."]

def merge_csv(dataframes):
    """
    Merges dataframes along the vertical axis
    
    Args:
        dataframes (``list`` of ``pandas.core.frame.DataFrame``): list of dataframes with same columns
    
    Returns:
        ``pandas.core.frame.DataFrame``: A merged dataframe resulting from 'stacking' all input dataframes

    """
    final_dataframe = pd.concat(dataframes, axis=0, join='inner').sort_index()
    return final_dataframe

def prune_images(force=False):
    """
    Prunes all Docker images named ``otter-grade``
    """
    if not force:
        sure = input("Are you sure you want to prune Otter's grading images? This action cannot be undone [y/N] ")
        sure = bool(re.match(r"ye?s?", sure, flags=re.IGNORECASE))
    else:
        sure = True
    
    if sure:        
        images = docker.images()

        for img in images:
            if any([t.startswith(OTTER_DOCKER_IMAGE_TAG + ":") for t in img.repo_tags]):
                img.remove(force=True)

def generate_hash(path):
    """
    Reads in a file and returns an MD5 hash of its contents.

    Args:
        path (``str``): path to the file that will be read in and hashed

    Returns:
        ``str``: the hash value of the file
    """
    zip_hash = ""
    m = md5()
    with open(path, "rb") as f:
        data = f.read() # read file in chunk and call update on each chunk if file is large.
        m.update(data)
        zip_hash = m.hexdigest()
    return zip_hash



"""Cell generators for R notebooks"""

import nbformat

from ..utils import lock


def gen_export_cells(nb_name, instruction_text):
    """
    Generates export cells that instruct the student the run a code cell calling 
    ``ottr::export`` to generate and download their submission. The Markdown cell contains:

    .. code-block:: markdown

        ## Submission
        
        Make sure you have run all cells in your notebook in order before running the cell below, so 
        that all images/graphs appear in the output. The cell below will generate a zipfile for you 
        to submit. **Please save before exporting!**

    Additional instructions can be appended to this cell by passing a string to ``instruction_text``.

    The code cell contains:

    .. code-block:: python

        # Save your notebook first, then run this cell to export your submission.
        ottr::export("path/to/notebook.ipynb")
    
    Args:
        instruction_text (``str``): extra instructions for students when exporting
        **kwargs: a catch-all so that the API matches 
            ``otter.assign.cell_generators.gen_export_cells``
    
    Returns:
        ``list`` of ``nbformat.NotebookNode``: generated export cells
    """
    instructions = nbformat.v4.new_markdown_cell()
    instructions.source = "## Submission\n\nMake sure you have run all cells in your notebook in order before " \
    "running the cell below, so that all images/graphs appear in the output. The cell below will generate " \
    "a zip file for you to submit. **Please save before exporting!**"
    
    if instruction_text:
        instructions.source += '\n\n' + instruction_text

    export = nbformat.v4.new_code_cell()
    source_lines = ["# Save your notebook first, then run this cell to export your submission."]
    source_lines.append(f'ottr::export("{nb_name}")')
    export.source = "\n".join(source_lines)

    lock(instructions)
    lock(export)

    return [instructions, export, nbformat.v4.new_markdown_cell(" ")]     # last cell is buffer



"""Output directory creation for Otter Assign"""

import nbformat
import os
import pathlib
import shutil
import tempfile
import warnings

from .constants import NB_VERSION
from .notebook_transformer import transform_notebook
from .plugins import replace_plugins_with_calls
from .solutions import overwrite_seed_vars, strip_ignored_lines, strip_solutions_and_output
from .tests import write_test
from .utils import get_notebook_language, patch_copytree, remove_cell_ids


def write_autograder_dir(nb_path, output_nb_path, assignment):
    """
    Converts a master notebook to a solutions notebook and writes this notebook to the output directory,
    copying support files and writing tests as needed.

    Args:
        nb_path (``pathlib.Path``): path to master notebook
        output_nb_path (``pathlib.Path``): path to output file
        assignment (``otter.assign.assignment.Assignment``): the assignment configurations
    """
    assignment.notebook_basename = os.path.basename(str(output_nb_path))

    with open(nb_path) as f:
        nb = nbformat.read(f, as_version=NB_VERSION)

    if assignment.lang is None:
        try:
            assignment.lang = get_notebook_language(nb)
        except KeyError:
            warnings.warn("Could not auto-parse kernelspec from notebook; assuming Python")
            assignment.lang = "python"

    transformed_nb, test_files = transform_notebook(nb, assignment)

    # replace plugins
    transformed_nb = replace_plugins_with_calls(transformed_nb)

    # update assignment.test_files for R notebooks
    assignment.test_files |= assignment.is_r

    output_dir = output_nb_path.parent
    tests_dir = output_dir / 'tests'
    if assignment.test_files:
        os.makedirs(tests_dir, exist_ok=True)

    if assignment.requirements:
        output_fn = ("requirements.txt", "requirements.R")[assignment.is_r]
        if isinstance(assignment.requirements, list):
            with open(str(output_dir / output_fn), "w+") as f:
                f.write("\n".join(assignment.requirements))
            assignment.requirements = str(output_dir / output_fn)
        else:
            shutil.copy(assignment.requirements, str(output_dir / output_fn))

    if assignment.environment:
        output_fn = "environment.yml"
        shutil.copy(assignment.environment, str(output_dir / output_fn))
    
    # strip out ignored lines
    transformed_nb = strip_ignored_lines(transformed_nb)

    # write tests
    test_ext = (".R", ".py")[assignment.is_python]
    for test_name, test_file in test_files.items():
        test_path = tests_dir / (test_name + test_ext) if assignment.test_files else test_name
        write_test(transformed_nb, test_path, test_file, use_file=assignment.test_files)

    # write a temp dir for otter generate tests
    if assignment.generate:
        assignment._temp_test_dir = pathlib.Path(tempfile.mkdtemp())
        for test_name, test_file in test_files.items():
            test_path = assignment._temp_test_dir / (test_name + test_ext)
            write_test(transformed_nb, test_path, test_file, use_file=True)

    # TODO: this is a bad practice and only a monkey-patch for #340. we shold do some better parsing
    # of the nbformat version info to determine if this is necessary.
    remove_cell_ids(transformed_nb)

    # write notebook
    nbformat.write(transformed_nb, str(output_nb_path))

    # copy files
    for file in assignment.files:

        # if a directory, copy the entire dir
        if os.path.isdir(file):
            shutil.copytree(file, str(output_dir / os.path.basename(file)))
            
        else:
            # check that file is in subdir
            assert os.path.abspath(nb_path.parent) in os.path.abspath(file), \
                f"{file} is not in a subdirectory of the master notebook directory"
            file_path = pathlib.Path(file).resolve()
            rel_path = file_path.parent.relative_to(nb_path.parent)
            os.makedirs(output_dir / rel_path, exist_ok=True)
            shutil.copy(file, str(output_dir / rel_path))


def write_student_dir(nb_name, autograder_dir, student_dir, assignment):
    """
    Copies the autograder (solutions) directory and removes extraneous files, strips solutions from
    the notebook, and removes hidden tests from the tests directory.

    Args:
        nb_name (``str``): the master notebook name
        autograder_dir (``pathlib.Path``): the path to the autograder directory
        student_dir (``pathlib.Path``): the path to the student directory
        assignment (``otter.assign.assignment.Assignment``): the assignment configurations
    """
    if assignment.is_r:
        from .r_adapter.tests import remove_hidden_tests_from_dir
    else:
        from .tests import remove_hidden_tests_from_dir

    # copy autograder dir
    with patch_copytree():
        shutil.copytree(autograder_dir, student_dir, copy_function=shutil.copy)

    # remove requirements from student dir if present
    output_fn = ("requirements.txt", "requirements.R")[assignment.is_r]
    requirements = str(student_dir / output_fn)
    if os.path.isfile(requirements):
        os.remove(requirements)

    # remove environment from student dir if present
    output_fn = "environment.yml"
    environment = str(student_dir / output_fn)
    if os.path.isfile(environment):
        os.remove(environment)

    # strip solutions from student version
    student_nb_path = student_dir / nb_name
    with open(student_nb_path) as f:
        nb = nbformat.read(f, as_version=NB_VERSION)

    nb = strip_solutions_and_output(nb)

    # overwrite seed variable
    if isinstance(assignment.seed, dict):
        nb = overwrite_seed_vars(nb, assignment.seed["variable"], assignment.seed["student_value"])

    # remove hidden tests from student directory
    remove_hidden_tests_from_dir(nb, student_dir / 'tests', assignment, use_files=assignment.test_files)

    with open(student_nb_path, "w") as f:
        nbformat.write(nb, f)


def write_output_directories(master_nb_path, result_dir, assignment):
    """
    Converts a master notebook to an autograder and student directory based on configurations in 
    ``assignment``.

    Args:
        master_nb_path (``nbformat.NotebookNode``): the master notebook path
        result_dir (``pathlib.Path``): path to the result directory
        assignment (``otter.assign.assignment.Assignment``): the assignment configurations

    Returns:
        ``pathlib.Path``: the absolute path to the written autograder notebook
    """
    # create directories
    autograder_dir = result_dir / 'autograder'
    student_dir = result_dir / 'student'
    shutil.rmtree(autograder_dir, ignore_errors=True)
    shutil.rmtree(student_dir, ignore_errors=True)
    os.makedirs(autograder_dir, exist_ok=True)

    # write autograder directory
    output_nb_path = autograder_dir / master_nb_path.name
    write_autograder_dir(master_nb_path, output_nb_path, assignment)

    # write student dir
    write_student_dir(master_nb_path.name, autograder_dir, student_dir, assignment)

    return os.path.abspath(output_nb_path)



"""
otter Python API
"""

import platform

from . import api
from .check import logs
from .check.notebook import Notebook
from .version import __version__

# whether Otter is running on Window
_WINDOWS = platform.system() == "Windows"



"""Convert a v0 formatted notebook to a v1 formatted notebook"""

import copy
import nbformat
import re

from ..assignment import is_assignment_cell
from ..constants import BEGIN_TEST_CONFIG_REGEX, BLOCK_QUOTE
from ..questions import is_question_cell, read_question_metadata
from ..solutions import is_markdown_solution_cell
from ..tests import is_test_cell
from ..utils import get_source, get_spec, is_markdown_cell

from ...blocks import BlockType


def generate_delim_cell(block_type, end=False):
    """
    Generates a block boundary cell for the specified block type.

    Args:
        block_type (``BlockType``): the block type to deliminate
        end (``bool``): wheter this is an end block cell

    Returns:
        ``nbformat.NotebookNode``: the generated boundary cell
    """
    source = f"# {'END' if end else 'BEGIN'} { block_type.value.upper() }"
    return nbformat.v4.new_raw_cell(source)


def transform_test_cell(cell):
    """
    Transform a test cell from v0 to v1 format.

    Args:
        cell (``nbformat.NotebookNode``): the cell to transform

    Returns:
        ``nbformat.NotebookNode``: the transformed cell
    """
    cell = copy.deepcopy(cell)
    source = get_source(cell)
    if re.match(BEGIN_TEST_CONFIG_REGEX, source[0]):
        return cell
    else:
        if "hidden" in source[0].lower():
            source = ["# HIDDEN"] + source[1:]
        else:
            source = source[1:]
    cell["source"] = "\n".join(source)
    return cell


def transform_assignment_cell(cell):
    """
    Transform an assignment config cell from v0 to v1 format.

    Args:
        cell (``nbformat.NotebookNode``): the cell to transform

    Returns:
        ``nbformat.NotebookNode``: the transformed cell
    """
    source = get_source(cell)
    begin_assignment_line = get_spec(source, "assignment")
    i, lines = begin_assignment_line + 1, []
    while source[i].strip() != BLOCK_QUOTE:
        lines.append(source[i])
        i = i + 1
    return nbformat.v4.new_raw_cell("\n".join(["# ASSIGNMENT CONFIG"] + lines))


def extract_question_metadata(cell):
    """
    Extract the question metadata lines from a question cell and return the original source lines
    stripped of the metadata block as well as the lines of YAML metadata.

    Args:
        cell (``nbformat.NotebookNode``): the cell to transform

    Returns:
        ``tuple[list[str], list[str]]``: the stripped question and the YAML lines
    """
    source = get_source(cell)
    begin_question_line = get_spec(source, "question")
    i, lines = begin_question_line + 1, []
    while source[i].strip() != BLOCK_QUOTE:
        lines.append(source[i])
        i = i + 1
    source = source[:begin_question_line-1] + source[i+1:]
    return source, lines


def get_transformed_cells(cells):
    """
    Transform a list of cells defining a v0-formatted notebook into a v1-formatted notebook.

    Args:
        cells (``list[nbformat.NotebookNode]``): the original notebook cells

    Returns:
        ``list[nbformat.NotebookNode]``: the transformed notebook cells
    """
    transformed_cells = []
    question_metadata, processed_solution, tests_started = False, False, False

    for cell in cells:

        # this is the prompt cell or if a manual question then the solution cell
        if question_metadata and not processed_solution:
            assert not is_question_cell(cell), f"Found question cell before end of previous question cell: {cell}"

            # if this isn't a MD solution cell but in a manual question, it has a Markdown prompt
            if question_metadata.get('manual', False) and is_markdown_cell(cell) and not is_markdown_solution_cell(cell):
                md_has_prompt = True
                transformed_cells.append(generate_delim_cell(BlockType.PROMPT))
                transformed_cells.append(cell)
                transformed_cells.append(generate_delim_cell(BlockType.PROMPT, end=True))
                continue

            # if this is a test cell, this question has no response cell for the students, so we don't
            # include it in the output notebook but we need a test file
            elif is_test_cell(cell):
                transformed_cells.append(generate_delim_cell(BlockType.TESTS))
                transformed_cells.append(transform_test_cell(cell))
                no_solution = True
                tests_started = True

            else:
                transformed_cells.append(generate_delim_cell(BlockType.SOLUTION))
                transformed_cells.append(cell)
                transformed_cells.append(generate_delim_cell(BlockType.SOLUTION, end=True))
            
            processed_solution = True

        # if this is a test cell, parse and add to test_cases
        elif question_metadata and processed_solution and is_test_cell(cell):
            if not tests_started:
                transformed_cells.append(generate_delim_cell(BlockType.TESTS))
                tests_started = True
            transformed_cells.append(transform_test_cell(cell))

        else:
            # the question is over -- we've seen the question and solution and any tests and now we 
            # need to get ready to process the next question which *could be this cell*
            if question_metadata and processed_solution:

                if tests_started:
                    transformed_cells.append(generate_delim_cell(BlockType.TESTS, end=True))
                transformed_cells.append(generate_delim_cell(BlockType.QUESTION, end=True))

                # reset vars
                question_metadata, processed_solution, tests_started = False, False, False

            if is_assignment_cell(cell):
                transformed_cells.append(transform_assignment_cell(cell))

            # if a question cell, parse metadata, comment out question metadata, and append to nb
            elif is_question_cell(cell):
                new_source, meta_lines = extract_question_metadata(cell)
                new_cell = generate_delim_cell(BlockType.QUESTION)
                source = get_source(new_cell)
                source.extend(meta_lines)
                new_cell["source"] =  "\n".join(source)

                transformed_cells.append(new_cell)
                transformed_cells.append(nbformat.v4.new_markdown_cell("\n".join(new_source)))

                question_metadata = read_question_metadata(cell)

            else:
                assert not is_test_cell(cell), f"Test outside of a question: {cell}"
                transformed_cells.append(cell)

    if tests_started:
        transformed_cells.append(generate_delim_cell(BlockType.TESTS, end=True))

    if question_metadata is not False:
        transformed_cells.append(generate_delim_cell(BlockType.QUESTION, end=True))

    return transformed_cells



"""
Solution removal for Otter Assign
"""

import copy
import re

from .utils import rmd_to_cells, collapse_empty_cells

from ..constants import MD_RESPONSE_CELL_SOURCE
from ..r_adapter.solutions import solution_assignment_sub, solution_line_sub
from ..utils import get_source


def is_markdown_solution_cell(cell):
    """
    Returns whether any line of the cell matches `<!-- BEGIN SOLUTION -->`
    
    Args:
        cell (``nbformat.NotebookNode``): notebook cell
    
    Returns:
        ``bool``: whether the current cell is a Markdown solution cell
    """
    if not cell.cell_type == 'markdown':
        return False
    source = get_source(cell)
    return source and any([re.match(r"\s*<!-- BEGIN SOLUTION -->\s*", l, flags=re.IGNORECASE) for l in source])


def overwrite_seed_vars(rmd_string, seed_variable, seed):
    """
    Overwrite any assignments of the variable named ``seed_variable`` with the value ``seed`` in a
    notebook.

    Args:
        rmd_string (``str``): the Rmd document as a string
        seed_variable (``str``): the variable to look for
        seed (``int``): the value to set for ``seed_variable``

    Returns:
        ``nbformat.NotebookNode``: the notebook with the variable substitutions made
    """
    lines = rmd_string.split("\n")
    in_code_block = False
    for i, line in enumerate(lines):
        if line.startswith("```{r"):
            in_code_block = True
        elif in_code_block and line.startswith("```"):
            in_code_block = False
        elif in_code_block:
            match = re.match(fr"(\s*){seed_variable}\s*(=|<-)\s*", line)
            if match:
                lines[i] = match.group(1) + f"{seed_variable} {match.group(2)} {seed}"
    return "\n".join(lines)


solution_assignment_regex = re.compile(r"(\s*[a-zA-Z0-9_. ]*(=|<-))(.*)[ ]?#[ ]?SOLUTION")
solution_line_regex = re.compile(r"(\s*)([^#\n]+)[ ]?#[ ]?SOLUTION")

begin_solution_regex = re.compile(r"(\s*)# BEGIN SOLUTION( NO PROMPT)?")
skip_suffixes = [
    '# SOLUTION NO PROMPT', '# BEGIN PROMPT', '# END PROMPT', '# SEED', '<!-- BEGIN PROMPT -->',
    '<!-- END PROMPT -->'
]

begin_md_solution_regex = re.compile(r"(\s*)<!--\s*BEGIN SOLUTION( NO PROMPT)?\s*-->")

SUBSTITUTIONS = [
    (solution_assignment_regex, solution_assignment_sub),
    (solution_line_regex, solution_line_sub),
]


def replace_solutions(lines):
    """
    Replaces solutions in ``lines``
    
    Args:
        lines (``list`` of ``str``): solutions as a list of strings

    Returns:
        ``list`` of ``str``: stripped version of lines without solutions
    """
    stripped = []
    solution = False
    for line in lines:

        # ...
        if any(line.endswith(s) for s in skip_suffixes):
            continue

        # ...
        if solution and not (line.endswith('# END SOLUTION') or line.rstrip().endswith("<!-- END SOLUTION -->")):
            continue

        # ...
        if line.endswith('# END SOLUTION') or line.rstrip().endswith("<!-- END SOLUTION -->"):
            assert solution, f"END SOLUTION without BEGIN SOLUTION in {lines}"
            solution = False
            continue

        # ...
        begin_solution = begin_solution_regex.match(line)
        if begin_solution:
            assert not solution, f"Nested BEGIN SOLUTION in {lines}"
            solution = True
            if not begin_solution.group(2):
                line = begin_solution.group(1) + '# YOUR CODE HERE'
            else:
                continue
        
        begin_solution = begin_md_solution_regex.match(line)
        if begin_solution:
            assert not solution, f"Nested BEGIN SOLUTION in {lines}"
            solution = True
            if not begin_solution.group(2):
                line = begin_solution.group(1) + MD_RESPONSE_CELL_SOURCE
            else:
                continue
        
        for exp, sub in SUBSTITUTIONS:
            m = exp.match(line)
            if m:
                line = sub(m)
        
        stripped.append(line)
    
    assert not solution, f"BEGIN SOLUTION without END SOLUTION in {lines}"
    
    return stripped


def strip_solutions_and_output(rmd_string):
    """
    Writes a Rmd file with solutions stripped
    
    Args:
        rmd_string (``str``): the Rmd document as a string
    """
    md_solutions = []
    cells = rmd_to_cells(rmd_string)
    for i, cell in enumerate(cells):
        cell = copy.deepcopy(cells[i])
        cell["source"] = '\n'.join(replace_solutions(get_source(cell)))
        cells[i] = cell
    collapse_empty_cells(cells)
    rmd_string = "\n".join([c.source for c in cells])    
    return rmd_string



"""Autograder runner for R assignments"""

import copy
import json
import nbformat
import os
import pickle
import shutil
import tempfile

from glob import glob
from nbconvert.exporters import ScriptExporter
from rpy2.robjects.packages import importr

from .abstract_runner import AbstractLanguageRunner
from ..utils import OtterRuntimeError
from ....export import export_notebook
from ....generate.token import APIClient
from ....test_files import GradingResults
from ....utils import chdir, get_source, knit_rmd_file


NBFORMAT_VERSION = 4
R_PACKAGES = {
    "knitr": importr("knitr"),
    "ottr": importr("ottr"),
}


class RRunner(AbstractLanguageRunner):

    subm_path_deletion_reauired = False
    """whether the submission path needs to be deleted (because it was created with tempfile)"""

    def filter_cells_with_syntax_errors(self, nb):
        """
        Filter out cells in an R notebook with syntax errors.
        """
        new_cells = []
        for cell in nb["cells"]:
            if cell["cell_type"] == "code":
                source = "\n".join(get_source(cell))
                valid_syntax = R_PACKAGES["ottr"].valid_syntax(source)[0]
                if valid_syntax:
                    new_cells.append(cell)
        nb = copy.deepcopy(nb)
        nb["cells"] = new_cells
        return nb

    def add_seeds_to_rmd_file(self, rmd_path):
        """
        Add intercell seeding to an Rmd file.
        """
        with open(rmd_path) as f:
            rmd = f.read()

        lines = rmd.split("\n")
        insertions = []
        for i, line in enumerate(lines):
            if line.startswith("```{r"):
                insertions.append(i)

        seed = f"set.seed({self.options['seed']})"
        if self.options["seed_variable"]:
            seed = f"{self.options['seed_variable']} = {self.options['seed']}"

        for i in insertions[::-1]:
            lines.insert(i + 1, seed)

        with open(rmd_path, "w") as f:
            f.write("\n".join(lines))

    def add_seed_to_script(self, script_path):
        """
        Add intercell seeding to an Rmd file.
        """
        with open(script_path) as f:
            script = f.read()

        script = f"set.seed({self.options['seed']})\n" + script

        with open(script_path, "w") as f:
            f.write(script)

    def resolve_submission_path(self):
        # create a temporary file at which to write a script if necessary
        _, script_path = tempfile.mkstemp(suffix=".R")

        # convert IPYNB files to Rmd files
        nbs = glob("*.ipynb")
        if len(nbs) > 1:
            raise OtterRuntimeError("More than one IPYNB file found in submission")

        elif len(nbs) == 1:
            nb_path = nbs[0]
            nb = nbformat.read(nb_path, as_version=NBFORMAT_VERSION)
            nb = self.filter_cells_with_syntax_errors(nb)

            # create the R script
            script, _ = ScriptExporter().from_notebook_node(nb)
            with open(script_path, "w") as f:
                f.write(script)

            self.subm_path_deletion_reauired = True
            return script_path

        # convert Rmd files to R files
        rmds = glob("*.Rmd")
        if len(rmds) > 1:
            raise OtterRuntimeError("More than one Rmd file found in submission")

        elif len(rmds) == 1:
            rmd_path = rmds[0]

            # add seeds
            if self.options["seed"] is not None:
                self.add_seeds_to_rmd_file(rmd_path)

            # create the R script
            rmd_path = os.path.abspath(rmd_path)
            R_PACKAGES["knitr"].purl(rmd_path, script_path)

            self.subm_path_deletion_reauired = True
            return script_path

        os.remove(script_path)

        # get the R script
        scripts = glob("*.[Rr]")
        if len(scripts) > 1:
            raise OtterRuntimeError("More than one R script found in submission")

        elif len(scripts) == 0:
            raise OtterRuntimeError("No gradable files found in submission")

        if self.options["seed"] is not None:
            self.add_seed_to_script(scripts[0]) 

        return scripts[0]

    def write_pdf(self):
        """
        Generate a PDF of a submission using the options in ``self.options`` and return the that to 
        the PDF.
        """
        try:
            nbs = glob("*.ipynb")
            if nbs:
                subm_path = nbs[0]
                ipynb = True

            else:
                rmds = glob("*.Rmd")
                if rmds:
                    subm_path = rmds[0]
                    ipynb = False

                else:
                    raise OtterRuntimeError("Could not find a file that can be converted to a PDF")

            pdf_path = os.path.splitext(subm_path)[0] + ".pdf"
            if ipynb:
                export_notebook(
                    subm_path, dest=pdf_path, filtering=self.options["filtering"], 
                    pagebreaks=self.options["pagebreaks"], exporter_type="latex")

            else:
                knit_rmd_file(subm_path, pdf_path)

        except Exception as e:
            print(f"\n\nError encountered while generating and submitting PDF:\n{e}")

        return pdf_path

    # TODO
    def submit_pdf(self, client, pdf_path):
        """
        Upload a PDF to a Gradescope assignment for manual grading.

        Args:
            client (``otter.generate.token.APIClient``): the Gradescope client
            pdf_path (``str``): path to the PDF
        """
        try:
            # get student email
            with open("../submission_metadata.json", encoding="utf-8") as f:
                metadata = json.load(f)

            student_emails = []
            for user in metadata["users"]:
                student_emails.append(user["email"])

            for student_email in student_emails:
                client.upload_pdf_submission(
                    self.options["course_id"], self.options["assignment_id"], student_email, pdf_path)

            print("\n\nSuccessfully uploaded submissions for: {}".format(", ".join(student_emails)))

        except Exception as e:
            print(f"\n\nError encountered while generating and submitting PDF:\n{e}")

    def run(self):
        os.environ["PATH"] = f"{self.options['miniconda_path']}/bin:" + os.environ.get("PATH")

        with chdir("./submission"):
            if self.options["token"] is not None:
                client = APIClient(token=self.options["token"])
                generate_pdf = True
                has_token = True

            else:
                generate_pdf = self.options["pdf"]
                has_token = False
                client = None

            subm_path = self.resolve_submission_path()
            ignore_errors = "FALSE" if self.options["debug"] else "TRUE"
            output = R_PACKAGES["ottr"].run_autograder(
                subm_path, ignore_errors = not self.options["debug"])[0]
            scores = GradingResults.from_ottr_json(output)

            if generate_pdf:
                pdf_path = self.write_pdf()

                if has_token:
                    self.submit_pdf(client, pdf_path)

        # delete the script if necessary
        if self.subm_path_deletion_reauired:
            os.remove(subm_path)
            self.subm_path_deletion_reauired = False

        return scores



#################################
##### Tests for otter check #####
#################################

import os
import unittest
import subprocess
import contextlib

from textwrap import dedent
from subprocess import PIPE
from glob import glob
from io import StringIO
from unittest import mock

from otter import Notebook
from otter.check import main as check

from . import TestCase

# parser = get_parser()

TEST_FILES_PATH = "test/test-check/"

class TestCheck(TestCase):

    def test_otter_check_script(self):
        """
        Checks that the script checker works
        """
        # run for each individual test
        for file in glob(TEST_FILES_PATH + "tests/*.py"):
            # capture stdout
            output = StringIO()
            with contextlib.redirect_stdout(output):

                # mock block_print otherwise it interferes with capture of stdout
                with mock.patch("otter.check.block_print"):
                    check(
                        TEST_FILES_PATH + "file0.py", 
                        question = os.path.split(file)[1][:-3],
                        tests_path = os.path.split(file)[0],
                    )

                    if os.path.split(file)[1] != "q2.py":
                        self.assertEqual(
                            output.getvalue().strip().split("\n")[-1].strip(), 
                            "All tests passed!", 
                            "Did not pass test at {}".format(file)
                        )

        # run the file for all questions
        output = StringIO()
        with contextlib.redirect_stdout(output):
            with mock.patch("otter.check.block_print"):
                check(
                    TEST_FILES_PATH + "file0.py", 
                    tests_path = os.path.split(file)[0],
                )
                self.assertEqual(
                    output.getvalue().strip(), 
                    dedent("""\
                        [0.         0.02002002 0.04004004 0.06006006 0.08008008]
                        q1 results: All test cases passed!
                        q2 results:
                            q2 - 1 result:
                                Trying:
                                    1 == 1
                                Expecting:
                                    False
                                **********************************************************************
                                Line 2, in q2 0
                                Failed example:
                                    1 == 1
                                Expected:
                                    False
                                Got:
                                    True

                            q2 - 2 result:
                                Test case passed!
                        q3 results: All test cases passed!
                        q4 results: All test cases passed!
                        q5 results: All test cases passed!"""), 
                    "Did not pass correct tests"
                )

    def test_otter_check_notebook(self):
        """
        Checks that the script checker works
        """
        # run for each individual test
        for file in glob(TEST_FILES_PATH + "tests/*.py"):
            # capture stdout
            output = StringIO()
            with contextlib.redirect_stdout(output):

                # mock block_print otherwise it interferes with capture of stdout
                with mock.patch("otter.check.block_print"):
                    check(
                        TEST_FILES_PATH + "test-nb.ipynb", 
                        question = os.path.split(file)[1][:-3],
                        tests_path = os.path.split(file)[0],
                    )

                    if os.path.split(file)[1] != "q2.py":
                        self.assertEqual(
                            output.getvalue().strip().split("\n")[-1].strip(), 
                            "All tests passed!", 
                            "Did not pass test at {}".format(file)
                        )

        # run the file for all questions
        output = StringIO()
        with contextlib.redirect_stdout(output):
            with mock.patch("otter.check.block_print"):
                check(
                    TEST_FILES_PATH + "test-nb.ipynb", 
                    tests_path = os.path.split(file)[0],
                )

                self.assertEqual(
                    output.getvalue().strip(), 
                    dedent("""\
                        [0.         0.02002002 0.04004004 0.06006006 0.08008008]
                        q1 results: All test cases passed!
                        q2 results:
                            q2 - 1 result:
                                Trying:
                                    1 == 1
                                Expecting:
                                    False
                                **********************************************************************
                                Line 2, in q2 0
                                Failed example:
                                    1 == 1
                                Expected:
                                    False
                                Got:
                                    True

                            q2 - 2 result:
                                Test case passed!
                        q3 results: All test cases passed!
                        q4 results: All test cases passed!
                        q5 results: All test cases passed!"""), 
                    "Did not pass correct tests"
                )
    
    @classmethod
    def tearDownClass(cls):
        if os.path.exists(".OTTER_LOG"):
            os.system("rm .OTTER_LOG")



"""Various utilities for Otter-Grader"""

import os
import sys
import pathlib
import random
import re
import string
import shutil
import tempfile

from contextlib import contextmanager, redirect_stdout
from IPython import get_ipython

@contextmanager
def block_print():
    """
    Context manager that disables printing to stdout
    """
    with open(os.devnull, "w") as f, redirect_stdout(f):
        yield


@contextmanager
def hide_outputs():
    """
    Context manager for hiding outputs from ``display()`` calls. IPython handles matplotlib outputs 
    specially, so those are supressed too.
    """
    ipy = get_ipython()
    if ipy is None:
        # Not running inside ipython!
        yield
        return
    old_formatters = ipy.display_formatter.formatters
    ipy.display_formatter.formatters = {}
    try:
        yield
    finally:
        ipy.display_formatter.formatters = old_formatters


def id_generator(size=6, chars=string.ascii_uppercase + string.digits):
    """
    Used to generate a dynamic variable name for grading functions

    This function generates a random name using the given length and character set.
    
    Args:
        size (``int``): length of output name
        chars (``str``, optional): set of characters used to create function name
    
    Returns:
        ``str``: randomized string name for grading function
    """
    return ''.join(random.choice(chars) for _ in range(size))


def get_variable_type(obj):
    """
    Returns the fully-qualified type string of an object ``obj``

    Args:
        obj (object): the object in question

    Returns:
        ``str``: the fully-qualified type string
    """
    return type(obj).__module__ + "." + type(obj).__name__


def get_relpath(src, dst):
    """
    Returns the relative path from ``src`` to ``dst``

    Args:
        src (``pathlib.Path``): the source directory
        dst (``pathlib.Path``): the destination directory

    Returns:
        ``pathlib.Path``: the relative path
    """
    # osrc = src
    ups = 0
    while True:
        try:
            dst.relative_to(src)
            break
        except ValueError:
            src = src.parent
            ups += 1
    return pathlib.Path(("../" * ups) / dst.relative_to(src))


@contextmanager
def chdir(new_dir):
    """
    Create a context with a different working directory, resetting the working directory on exit.

    Args:
        new_dir (path-like): the directory for the context
    """
    curr_dir = os.getcwd()
    os.chdir(new_dir)

    try: 
        yield

    finally: 
        os.chdir(curr_dir)


def get_source(cell):
    """
    Returns the source code of a cell in a way that works for both nbformat and JSON
    
    Args:
        cell (``nbformat.NotebookNode``): notebook cell
    
    Returns:
        ``list`` of ``str``: each line of the cell source stripped of ending line breaks
    """
    source = cell.source
    if isinstance(source, str):
        return re.split("\r?\n", source)
    elif isinstance(source, list):
        return [line.strip("\r\n") for line in source]
    raise ValueError(f'unknown source type: {type(source)}')


@contextmanager
def nullcontext():
    """
    Yields an empty context. Added because ``contextlib.nullcontext`` was added in Python 3.7, so
    earlier versions of Python require this patch.
    """
    yield


@contextmanager
def load_default_file(provided_fn, default_fn, default_disabled=False):
    """
    Reads the contents of a file with an optional default path. If ``proivided_fn`` is not specified
    and ``expected_fn`` is a existing file path, the contents of ``expected_fn`` are read in place
    of ``provided_fn``. The use of ``expected_fn`` can be disabled by setting ``default_disabled``
    to ``True``.
    """
    if provided_fn is None and os.path.isfile(default_fn) and not default_disabled:
        provided_fn = default_fn
    
    if provided_fn is not None:
        assert os.path.isfile(provided_fn), f"Could not find specified file: {provided_fn}"
        with open(provided_fn) as f:
            yield f.read()
    else:
        yield None


def print_full_width(char, mid_text="", whitespace=" ", ret_str=False, **kwargs):
    """
    Prints a character at the full terminal width. If ``mid_text`` is supplied, this text is printed
    in the middle of the terminal, surrounded by ``whitespace``. Additional kwargs passed to 
    ``print``.

    If ``ret_str`` is true, the string is returned; if not, it is printed directly to the console.
    """
    cols, rows = shutil.get_terminal_size()

    if mid_text:
        left = cols - len(mid_text) - 2 * len(whitespace)
        if left <= 0:
            left = 2
        l, r = left // 2, left // 2
        if left % 2 == 1:
            r += 1
        
        out = char * l + whitespace + mid_text + whitespace + char * r

    else:
        out = char * cols

    if ret_str:
        return out

    print(out, **kwargs)


def convert_config_description_dict(configs, for_docs=False):
    """
    Recursively converts a documented list of dictionary configurations into a dictionary with the 
    default values loaded.

    Expects a list of the form:

    .. code-block:: python

        [
            {
                "key": "config_name",
                "description": "a description of the config for documentation",
                "default": True,  # the default config value
            },
            {
                "key": "required_config_name",
                "description": "a description of the config for documentation",
                "required": True,  # indicates that this must be user-specified and has no default
            },
            {
                "key": "nested_config_name",
                "description": "a description of the config for documentation",
                "default": None,
                "subkeys": [  # note that a list is used for nested dict configs
                    {
                        "key": "subconfig",
                        "description": "a nested key",
                        "default": None,
                    }
                ],
            },
        ]

    The list above gets converted to a dictionary mapping each ``key`` to each ``default``:

    .. code-block:: python

        {
            "config_name": True,
            "nested_config_name": None,
        }

    If ``for_docs`` is true, then any specified subkeys are set as the mapped value in the dictionary,
    and if the config is required, its default is set to ``None``.

    .. code-block:: python

        {
            "config_name": True,
            "required_config_name": None,
            "nested_config_name": {
                "subconfig": None,
            },
        }

    Any configurations with ``default`` unspecified have their default value set to ``None``. Any
    configurations marked with ``"required": True`` are not included in the output dict (so that
    they raise a ``KeyError`` if unspecified).

    Args:
        configs (``list[dict[str,object]]``): the configurations with the structure defined above

    Returns:
        ``dict[str,object]``: a dictionary mapping keys to default values
    """
    res = {}
    for d in configs:
        default = d.get("default")
        subkeys = d.get("subkeys")
        if isinstance(default, list) and len(default) > 0 and \
                all(isinstance(e, dict) for e in default):
            default = convert_config_description_dict(default)
        elif for_docs and isinstance(subkeys, list) and len(subkeys) > 0 and \
                all(isinstance(e, dict) for e in subkeys):
            default = convert_config_description_dict(subkeys)
        if not d.get("required", False) or for_docs:
            res[d["key"]] = default
    return res


def assert_path_exists(path_tuples):
    """
    Ensure that a series of file paths exist and are of a specific type, or raise a ``ValueError``.
    
    Elements of ``path_tuples`` should be 2-tuples where the first element is a string representing 
    the file path and the second element is ``True`` if the path should be a directory, ``False`` if 
    it should be a file, and ``None`` if it doesn't matter.

    Args:
        path_tuples (``list[tuple[str, bool]]``): the list of paths as described above

    Raises:
        ``ValueError``: if the path does not exist or it is not of the correct type
    """
    for path, is_dir in path_tuples:
        if not os.path.exists(path):
            raise ValueError(f"Path {path} does not exist")
        if is_dir and not os.path.isdir(path):
            raise ValueError(f"Path {path} is not a directory")
        if is_dir is False and not os.path.isfile(path):
            raise ValueError(f"Path {path} is not a file")


def knit_rmd_file(rmd_path, pdf_path):
    """
    Use ``rpy2`` and ``rmarkdown::render`` to knit an RMarkdown file to a PDF, allowing errors.

    Args:
        rmd_path (``str``): the path to the Rmd file
        pdf_path (``str``): the path at which to write the PDF
    """
    from rpy2.robjects.packages import importr

    with tempfile.NamedTemporaryFile(mode="w", suffix=".Rmd",) as ntf:
        with open(rmd_path) as f:
            contents = f.read()

        contents = "```{r cache = F, include = F}\nknitr::opts_chunk$set(error = TRUE)\n```\n" + contents
        ntf.write(contents)
        ntf.seek(0)

        pdf_path = os.path.abspath(pdf_path)
        rmarkdown = importr("rmarkdown")
        rmarkdown.render(ntf.name, "pdf_document", pdf_path)



"""IPython notebook PDF Exporter forked from nb2pdf and gsExport"""

import nbformat
import nbconvert
import os
import pkg_resources
import warnings

from .exporters import get_exporter


def export_notebook(nb_path, dest=None, debug=False, exporter_type=None, **kwargs):
    """
    Exports a notebook file at ``nb_path`` to a PDF with optional filtering and pagebreaks. Accepts
    other ``kwargs`` passed to the exporter class's ``convert_notebook`` class method.

    Args:
        nb_path (``str``): path to notebook
        dest (``str``, optional): path to write PDF
        debug (``bool``, optional): whether to run export in debug mode
        exporter_type (``str``, optional): the type of exporter to use; one of ``['html', 'latex']``
        **kwargs: additional configurations passed to exporter
    """
    # notebook = load_notebook(nb_path, filtering=filtering, pagebreaks=pagebreaks)

    if dest is not None:
        pdf_name = dest
    else:
        pdf_name = os.path.splitext(nb_path)[0] + ".pdf"
        
    # notebook_to_pdf(notebook, pdf_name, save_tex=save_tex, debug=debug)
    Exporter = get_exporter(exporter_type=exporter_type)
    Exporter.convert_notebook(nb_path, pdf_name, debug=debug, **kwargs)


def main(src, *, dest=None, exporter=None, filtering=False, pagebreaks=False, save=False, debug=False):
    """
    Runs Otter Export

    Args:
        src (``str``): path to source notebook
        dest (``Optional[str]``): path at which to write PDF
        exporter (``Optional[str]``): exporter name
        filtering (``bool``): whether to filter cells using HTML comments
        pagebreaks (``bool``): whether to pagebreak between filtered regions; ignored if ``filtering``
            is ``False``
        save (``bool``): whether to save any intermediate files (e.g. ``.tex``, ``.html``)
        debug (``bool``): whether to run in debug mode (print full error messages)
        **kwargs: ignored kwargs (a remnant of how the argument parser is built)
    """
    export_notebook(
        src,
        dest = dest,
        exporter_type = exporter,
        filtering = filtering,
        pagebreaks = pagebreaks,
        save_tex = save,
        save_html = save,
        debug = debug
    )



"""A class to collect test file data from R autograding with Ottr"""

from .abstract_test import TestFile


class OttrTestFile(TestFile):

    def run(self, global_environment):
        raise NotImplementedError("Ottr test files cannot be run from Python")
    
    @classmethod
    def from_file(cls, path):
        raise NotImplementedError("Cannot create Ottr test files from a file in Python")



"""
Execution and grading internals for Otter-Grader
"""

import json
import inspect
import itertools
import nbformat

from IPython import get_ipython

from .execute_log import execute_log
from .execute_notebook import execute_notebook
from .transforms import filter_ignored_cells, script_to_notebook

from ..test_files import GradingResults, NotebookMetadataOKTestFile, OKTestFile
from ..utils import id_generator


NBFORMAT_VERSION = 4


def check(nb_or_test_path, test_name=None, global_env=None):
    """
    Checks a global environment against given test file. If global_env is ``None``, the global 
    environment of the calling frame is used; i.e., the following two calls are equivalent:

    .. code-block:: python

        check('tests/q1.py')
        check('tests/q1.py', globals())
    
    Args:
        nb_or_test_path (``str``): path to test file or notebook
        test_name (``str``, optional): the name of the test if a notebook metadata test
        global_env (``dict``, optional): a global environment resulting from the execution 
            of a python script or notebook

    Returns:
        ``otter.test_files.abstract_test.TestFile``: result of running the tests in the 
        given global environment
    """
    if test_name is None:
        test = OKTestFile.from_file(nb_or_test_path)
    else:
        test = NotebookMetadataOKTestFile.from_file(nb_or_test_path, test_name)

    if global_env is None:
        # Get the global env of our callers - one level below us in the stack
        # The grade method should only be called directly from user / notebook
        # code. If some other method is calling it, it should also use the
        # inspect trick to pass in its parents' global env.
        global_env = inspect.currentframe().f_back.f_globals

    test.run(global_env)

    return test


def grade_notebook(submission_path, *, tests_glob=None, name=None, ignore_errors=True, script=False, 
    cwd=None, test_dir=None, seed=None, seed_variable=None, log=None, variables=None, 
    plugin_collection=None):
    """
    Grade an assignment file and return grade information

    Args:
        submission_path (``str``): path to a single notebook or Python script
        tests_glob (``list`` of ``str``, optional): paths to test files to run
        name (``str``, optional): initial environment name
        ignore_errors (``bool``, optional): whether errors in execution should be ignored
        script (``bool``, optional): whether the ``submission_path`` is a Python script
        cwd (``str``, optional): working directory of execution to be appended to ``sys.path`` in 
            grading environment
        test_dir (``str``, optional): path to directory of tests in grading environment
        seed (``int``, optional): random seed for intercell seeding
        seed_variable (``str``, optional): a variable name to override with the seed
        log (``otter.check.logs.Log``, optional): log from which to grade questions
        variables (``dict``, optional): map of variable names -> type string to check type of deserialized
            object to prevent arbitrary code from being put into the environment; ignored if log is ``None``
        plugin_collection (``otter.plugins.PluginCollection``, optional): a set of plugins to run on
            this assignment during execution and grading

    Returns:
        ``otter.test_files.GradingResults``: the results of grading
    """
    if not script:
        nb = nbformat.read(submission_path, as_version=NBFORMAT_VERSION)

    else:
        with open(submission_path) as f:
            nb = f.read()

        nb = script_to_notebook(nb)

    if plugin_collection is not None:
        nb = plugin_collection.before_execution(nb)

    # remove any ignored cells from the notebook
    if not script:
        nb = filter_ignored_cells(nb)

    secret = id_generator()
    results_array = "check_results_{}".format(secret)
    initial_env = {
        results_array: []
    }

    if name:
        initial_env["__name__"] = name

    if log is not None:
        global_env = execute_log(
            nb, log, secret, initial_env, ignore_errors=ignore_errors, cwd=cwd, test_dir=test_dir, 
            variables=variables)

    else:
        global_env = execute_notebook(
            nb, results_array, initial_env, ignore_errors=ignore_errors, cwd=cwd, test_dir=test_dir, 
            seed=seed, seed_variable=seed_variable)

    if plugin_collection is not None:
        plugin_collection.run("after_execution", global_env)

    tests_run = global_env[results_array]

    # Check for tests which were not included in the notebook and specified by tests_globs
    # Allows instructors to run notebooks with additional tests not accessible to user
    if tests_glob:
        # unpack list of paths into a single list
        tested_set = [test.path for test in tests_run]
        extra_tests = []
        for t in sorted(tests_glob):
            include = True
            for tested in tested_set:
                if tested in t or t in tested:     # e.g. if 'tests/q1.py' is in /srv/repo/lab01/tests/q1.py
                    include = False

            if include:
                extra_tests.append(OKTestFile.from_file(t))
                extra_tests[-1].run(global_env)

        tests_run += extra_tests

    results = GradingResults(tests_run)

    if plugin_collection is not None:
        plugin_collection.run("after_grading", results)
    
    return results



##################################
##### Tests for Otter Export #####
##################################

# NOTES:
# - tests do not check for PDF equality

import os
import unittest
import subprocess
import json
import re
import contextlib
import nbconvert
import filecmp
import nbformat

from io import StringIO
from unittest import mock
from subprocess import Popen, PIPE
from glob import glob
from textwrap import dedent

from otter.export import export_notebook
from otter.export import main as export
from otter.export.exporters.base_exporter import BaseExporter

from . import TestCase


TEST_FILES_PATH = "test/test-export/"


class TestExport(TestCase):

    @staticmethod
    def run_export(notebook_stem, **kwargs):
        export(TEST_FILES_PATH + notebook_stem + ".ipynb", **kwargs)

    def test_success_HTML(self):
        """
        Tests a successful export with filtering and no pagebreaks
        """
        test_file = "successful-html-test"
        self.run_export(test_file, filtering=True, save=True, exporter="latex")

        # check existence of pdf and tex
        self.assertTrue(os.path.isfile(TEST_FILES_PATH + test_file + ".pdf"))
        self.assertTrue(os.path.isfile(TEST_FILES_PATH + test_file + ".tex"))

        # cleanup
        cleanup_command = ["rm", TEST_FILES_PATH + test_file + ".pdf", TEST_FILES_PATH + test_file + ".tex"]
        cleanup = subprocess.run(cleanup_command, stdout=PIPE, stderr=PIPE)
        self.assertEqual(cleanup.returncode, 0,"Error in cleanup: " + str(cleanup.stderr))
        
    def test_success_pagebreak(self):
        """
        Tests a successful filter with pagebreaks
        """
        test_file = "success-pagebreak-test"
        self.run_export(test_file, filtering=True, exporter="latex", pagebreaks=True, save=True)

        # check existence of pdf and tex
        self.assertTrue(os.path.isfile(TEST_FILES_PATH + test_file + ".pdf"))
        self.assertTrue(os.path.isfile(TEST_FILES_PATH + test_file + ".tex"))

        # cleanup
        cleanup_command = ["rm", TEST_FILES_PATH + test_file + ".pdf", TEST_FILES_PATH + test_file + ".tex"]
        cleanup = subprocess.run(cleanup_command, stdout=PIPE, stderr=PIPE)
        self.assertEqual(cleanup.returncode, 0,"Error in cleanup: " + str(cleanup.stderr))

    def test_no_close(self):
        """
        Tests a filtered export without a closing comment
        """
        test_file = "no-close-tag-test"
        self.run_export(test_file, filtering=True, exporter="latex", pagebreaks=True, save=True)

        # check existence of pdf and tex
        self.assertTrue(os.path.isfile(TEST_FILES_PATH + test_file + ".pdf"))
        self.assertTrue(os.path.isfile(TEST_FILES_PATH + test_file + ".tex"))

        # cleanup
        cleanup_command = ["rm", TEST_FILES_PATH + test_file + ".pdf", TEST_FILES_PATH + test_file + ".tex"]
        cleanup = subprocess.run(cleanup_command, stdout=PIPE, stderr=PIPE)
        self.assertEqual(cleanup.returncode, 0,"Error in cleanup:" + str(cleanup.stderr))

    def test_load_notebook(self):
        """
        Tests a successful load_notebook
        """
        test_file = "successful-html-test"
        node = BaseExporter.load_notebook(TEST_FILES_PATH + test_file + ".ipynb", filtering=True)

        nbformat.write(node, TEST_FILES_PATH + test_file)

        # check existence of file
        self.assertTrue(os.path.isfile(TEST_FILES_PATH + test_file))
        self.assertTrue(filecmp.cmp(TEST_FILES_PATH + test_file, TEST_FILES_PATH + "correct/" + test_file))
        
        # cleanup
        cleanup_command = ["rm", TEST_FILES_PATH + test_file]
        cleanup = subprocess.run(cleanup_command, stdout=PIPE, stderr=PIPE)
        self.assertEqual(cleanup.returncode, 0,"Error in cleanup: " + str(cleanup.stderr))



"""Assignment creation tool for Otter-Grader"""

import json
import os
import pathlib
import warnings

from .assignment import Assignment
from .utils import run_tests, write_otter_config_file, run_generate_autograder

from ..export import export_notebook
from ..export.exporters import WkhtmltopdfNotFoundError
from ..plugins import PluginCollection
from ..utils import block_print, get_relpath, knit_rmd_file


def main(master, result, *, no_pdfs=False, no_run_tests=False, username=None, password=None, 
         debug=False, v1=False):
    """
    Runs Otter Assign on a master notebook.
    
    Args:
        master (``str``): path to master notebook
        result (``str``): path to result directory
        no_pdfs (``bool``): whether to ignore any configurations indicating PDF generation for this run
        no_run_tests (``bool``): prevents Otter tests from being automatically run on the solutions 
            notebook
        username (``str``): a username for Gradescope for generating a token
        password (``str``): a password for Gradescope for generating a token
        debug (``bool``): whether to run in debug mode (without ignoring errors during testing)
        v1 (``bool``): whether to use Otter Assign Format v1 instead of v0
    """
    if not v1:
        warnings.warn(
            "Otter Assign format v0 will be deprecated in Otter v4 and removed in a later release.",
            FutureWarning)
            
        from .v0 import main as v0_main
        return v0_main(master, result, no_pdfs=no_pdfs, no_run_tests=no_run_tests, username=username, 
            password=password, debug=debug)

    master, result = pathlib.Path(os.path.abspath(master)), pathlib.Path(os.path.abspath(result))
    print("Generating views...")

    assignment = Assignment()

    result = get_relpath(master.parent, result)
    orig_dir = os.getcwd()
    os.chdir(master.parent)
    
    assignment.master, assignment.result = master, result

    if assignment.is_rmd:
        from .rmarkdown_adapter.output import write_output_directories
    else:
        from .output import write_output_directories

    try:
        output_nb_path = write_output_directories(master, result, assignment)

        # update seed variables
        if isinstance(assignment.seed, dict):
            if assignment.generate:
                if assignment.generate is True:
                    assignment.generate = {}
                assignment.generate["seed"] = assignment.seed["autograder_value"]
                assignment.generate["seed_variable"] = assignment.seed["variable"]

        # check that we have a seed if needed
        if assignment.seed_required:
            generate_args = assignment.generate
            if generate_args is True:
                generate_args = {'seed': None}
            assert not generate_args or generate_args.get('seed', None) is not None or \
                not assignment.is_python, "Seeding cell found but no seed provided"

        plugins = assignment.plugins
        if plugins:
            pc = PluginCollection(plugins, "", {})
            pc.run("during_assign", assignment)
            if assignment.generate is True:
                assignment.generate = {"plugins": []}
            if assignment.generate is not False:
                if not assignment.generate.get("plugins"):
                    assignment.generate["plugins"] = []
                assignment.generate["plugins"].extend(plugins)
        else:
            pc = None

        # generate Gradescope autograder zipfile
        if assignment.generate:
            print("Generating autograder zipfile...")
            run_generate_autograder(result, assignment, username, password, plugin_collection=pc)
        
        # generate PDF of solutions
        if assignment.solutions_pdf and not no_pdfs:
            print("Generating solutions PDF...")
            filtering = assignment.solutions_pdf == 'filtered'

            src = os.path.abspath(str(result / 'autograder' / master.name))
            dst = os.path.abspath(str(result / 'autograder' / (master.stem + '-sol.pdf')))
        
            if not assignment.is_rmd:
                try:
                    export_notebook(
                        src,
                        dest=dst,
                        filtering=filtering,
                        pagebreaks=filtering,
                        exporter_type="html",
                    )
                except WkhtmltopdfNotFoundError:
                    export_notebook(
                        src,
                        dest=dst,
                        filtering=filtering,
                        pagebreaks=filtering,
                    )

            else:
                if filtering:
                    raise ValueError("Filtering is not supported with RMarkdown assignments")

                knit_rmd_file(src, dst)

        # generate a tempalte PDF for Gradescope
        if assignment.template_pdf and not no_pdfs:
            print("Generating template PDF...")
            
            src = os.path.abspath(str(result / 'autograder' / master.name))
            dst = os.path.abspath(str(result / 'autograder' / (master.stem + '-template.pdf')))

            if not assignment.is_rmd:
                export_notebook(
                    src,
                    dest=dst, 
                    filtering=True, 
                    pagebreaks=True, 
                    exporter_type="latex",
                )

            else:
                raise ValueError(f"Filtering is not supported with RMarkdown assignments; use " + \
                    "solutions_pdf to generate a Gradescope template instead.")

        # generate the .otter file if needed
        if not assignment.is_rmd and assignment.save_environment:
            if assignment.is_r:
                warnings.warn(
                    "Otter Service and serialized environments are unsupported with R, "
                    "configurations ignored"
                )
            else:
                write_otter_config_file(master, result, assignment)

        # run tests on autograder notebook
        if assignment.run_tests and not no_run_tests and assignment.is_python:
            print("Running tests...")
            # with block_print():
            if isinstance(assignment.generate, bool):
                seed = None
            else:
                seed = assignment.generate.get('seed', None)
            
            if assignment._otter_config is not None:
                test_pc = PluginCollection(assignment._otter_config.get("plugins", []), output_nb_path, {})

            else:
                test_pc = pc

            run_tests(result / 'autograder' / master.name, debug=debug, seed=seed, plugin_collection=test_pc)
            print("All tests passed!")

    finally:
        os.chdir(orig_dir)



"""
Output writing for Otter Assign
"""

import os
import shutil
import pathlib
import warnings
import nbformat

from .notebook_transformer import transform_notebook
from .solutions import overwrite_seed_vars, strip_solutions_and_output

from ..r_adapter.tests import remove_hidden_tests_from_dir
from ..tests import write_test

def write_autograder_dir(rmd_path, output_rmd_path, assignment):
    """
    Converts a master Rmd file to a solutions Rmd file and writes this file to the output directory,
    copying support files and writing tests as needed.

    Args:
        rmd_path (``pathlib.Path``): path to master Rmd file
        output_rmd_path (``pathlib.Path``): path to output Rmd file
        assignment (``otter.assign.assignment.Assignment``): the assignment configurations
    """
    with open(rmd_path) as f:
        rmd_string = f.read()

    if assignment.lang is None:
        assignment.lang = "r"

    output_dir = output_rmd_path.parent
    tests_dir = output_dir / 'tests'
    os.makedirs(tests_dir, exist_ok=True)

    requirements = assignment.requirements
    if requirements is None and os.path.isfile("requirements.R"):
        requirements = "requirements.R"
    if requirements:
        assert os.path.isfile(requirements), f"Requirements file {requirements} not found"
        assignment.requirements = requirements

    if assignment.requirements:
        shutil.copy(requirements, str(output_dir / 'requirements.R'))


    environment = assignment.environment
    if environment is None and os.path.isfile("environment.yml"):
        environment = "environment.yml"
    if environment:
        assert os.path.isfile(environment), f"Environment file {environment} not found"
        assignment.environment = environment

    if assignment.environment:
        shutil.copy(environment, str(output_dir / 'environment.yml'))

    transformed_rmd_string, test_files = transform_notebook(rmd_string, assignment)

    # write notebook
    with open(output_rmd_path, "w+") as f:
        f.write(transformed_rmd_string)

    # write tests
    test_ext =".R"
    for test_name, test_file in test_files.items():
        write_test({}, tests_dir / (test_name + test_ext), test_file, use_file=True)

    # copy files
    for file in assignment.files:

        # if a directory, copy the entire dir
        if os.path.isdir(file):
            shutil.copytree(file, str(output_dir / os.path.basename(file)))
            
        else:
            # check that file is in subdir
            assert os.path.abspath(rmd_path.parent) in os.path.abspath(file), \
                f"{file} is not in a subdirectory of the master notebook directory"
            file_path = pathlib.Path(file).resolve()
            rel_path = file_path.parent.relative_to(rmd_path.parent)
            os.makedirs(output_dir / rel_path, exist_ok=True)
            shutil.copy(file, str(output_dir / rel_path))

def write_student_dir(rmd_name, autograder_dir, student_dir, assignment):
    """
    Copies the autograder (solutions) directory and removes extraneous files, strips solutions from
    the Rmd file, and removes hidden tests from the tests directory.

    Args:
        rmd_name (``str``): the master Rmd file name
        autograder_dir (``pathlib.Path``): the path to the autograder directory
        student_dir (``pathlib.Path``): the path to the student directory
        assignment (``otter.assign.assignment.Assignment``): the assignment configurations
    """
    # copy autograder dir
    shutil.copytree(autograder_dir, student_dir)

    # remove requirements from student dir if present
    requirements = str(student_dir / 'requirements.R')
    if os.path.isfile(requirements):
        os.remove(requirements)

    # remove environment from student dir if present
    environment = str(student_dir / 'environment.yml')
    if os.path.isfile(environment):
        os.remove(environment)

    # strip solutions from student version
    student_rmd_path = student_dir / rmd_name
    with open(student_rmd_path) as f:
        # nb = nbformat.read(f, as_version=NB_VERSION)
        rmd_string = f.read()

    rmd_string = strip_solutions_and_output(rmd_string)

    # overwrite seed variable
    if isinstance(assignment.seed, dict):
        rmd_string = overwrite_seed_vars(
            rmd_string, assignment.seed["variable"], assignment.seed["student_value"])

    with open(student_rmd_path, "w") as f:
        f.write(rmd_string)

    # remove hidden tests from student directory
    remove_hidden_tests_from_dir({}, student_dir / 'tests', assignment)

def write_output_directories(master_rmd_path, result_dir, assignment):
    """
    Converts a master Rmd file to an autograder and student directory based on configurations in 
    ``assignment``.

    Args:
        master_rmd_path (``pathlib.Path``): the master Rmd path
        result_dir (``pathlib.Path``): path to the result directory
        assignment (``otter.assign.assignment.Assignment``): the assignment configurations
    """
    # create directories
    autograder_dir = result_dir / 'autograder'
    student_dir = result_dir / 'student'
    shutil.rmtree(autograder_dir, ignore_errors=True)
    shutil.rmtree(student_dir, ignore_errors=True)
    os.makedirs(autograder_dir, exist_ok=True)

    # write autograder directory
    output_rmd_path = autograder_dir / master_rmd_path.name
    write_autograder_dir(master_rmd_path, output_rmd_path, assignment)

    # write student dir
    write_student_dir(master_rmd_path.name, autograder_dir, student_dir, assignment)



####################################
##### Tests for otter generate #####
####################################

import json
import os
import shutil
import subprocess
import unittest

from glob import glob
from subprocess import PIPE
from unittest import mock

from otter.generate import main as generate

from .. import TestCase


TEST_FILES_PATH = "test/test_generate/test-autograder/"


class TestAutograder(TestCase):

    def test_autograder(self):
        """
        Check that the correct zipfile is created by gs_generator.py
        """
        # create the zipfile
        generate(
            tests_path = TEST_FILES_PATH + "tests",
            output_dir = TEST_FILES_PATH,
            requirements = TEST_FILES_PATH + "requirements.txt",
            files = [TEST_FILES_PATH + "data/test-df.csv"],
            no_environment = True,  # don't use the environment.yml in the root of the repo
        )

        with self.unzip_to_temp(TEST_FILES_PATH + "autograder.zip", delete=True) as unzipped_dir:
            self.assertDirsEqual(unzipped_dir, TEST_FILES_PATH + "autograder-correct")

    def test_autograder_with_token(self):
        """
        Checks otter assign with token specified instead of username and password.
        """
        # create the zipfile
        with mock.patch("otter.generate.APIClient") as mocked_client:
            generate(
                tests_path = TEST_FILES_PATH + "tests",
                output_dir = TEST_FILES_PATH,
                requirements = TEST_FILES_PATH + "requirements.txt",
                config = TEST_FILES_PATH + "otter_config.json",
                files = [TEST_FILES_PATH + "data/test-df.csv"],
                no_environment = True,  # don't use the environment.yml in the root of the repo
            )
            mocked_client.assert_not_called()
    
        with self.unzip_to_temp(TEST_FILES_PATH + "autograder.zip", delete=True) as unzipped_dir:
            self.assertDirsEqual(unzipped_dir, TEST_FILES_PATH + "autograder-token-correct")

    def test_custom_env(self):
        """
        Check that a custom environment.yml is correctly read and modified
        """
        # create the zipfile
        generate(
            tests_path = TEST_FILES_PATH + "tests",
            output_dir = TEST_FILES_PATH,
            requirements = TEST_FILES_PATH + "requirements.txt",
            environment = TEST_FILES_PATH + "environment.yml",
            files = [TEST_FILES_PATH + "data/test-df.csv"],
        )

        with self.unzip_to_temp(TEST_FILES_PATH + "autograder.zip", delete=True) as unzipped_dir:
            self.assertDirsEqual(unzipped_dir, TEST_FILES_PATH + "autograder-custom-env")



"""
Abstract base classes for working with test files and classes to represent collections of test and 
their results
"""

from abc import ABC, abstractmethod
from collections import namedtuple
from textwrap import dedent, indent
from typing import Tuple, List, Dict, Any
from jinja2 import Template
from pygments import highlight
from pygments.lexers import PythonConsoleLexer
from pygments.formatters import HtmlFormatter


# class for storing the test cases themselves
#   - body is the string that gets run for the test
#   - hidden is the visibility of the test case
#   - points is the number of points this test case is worth
TestCase = namedtuple("TestCase", ["name", "body", "hidden", "points", "success_message", "failure_message"])


# class for storing the results of a single test _case_ (within a test file)
#   - message should be a string to print out to the student (ignored if passed is True)
#   - passed is whether the test case passed
#   - hidden is the visibility of the test case
TestCaseResult = namedtuple("TestCaseResult", ["test_case", "message", "passed"])


class TestFile(ABC):
    """
    A (abstract) single test file for Otter. This ABC defines how test results are represented and sets
    up the instance variables tracked by tests. Subclasses should implement the abstract class method
    ``from_file`` and the abstract instance method ``run``.

    Args:
        name (``str``): the name of test file
        path (``str``): the path to the test file
        test_cases (``list`` of ``TestCase``): a list of parsed tests to be run
        all_or_nothing (``bool``, optional): whether the test should be graded all-or-nothing across
            cases

    Attributes:
        name (``str``): the name of test file
        path (``str``): the path to the test file
        test_cases (``list`` of ``TestCase``): a list of parsed tests to be run
        all_or_nothing (``bool``): whether the test should be graded all-or-nothing across
            cases
        test_case_results (``list`` of ``TestCaseResult``): a list of results for the test cases in
            ``test_cases``
    """

    def _repr_html_(self):
        if self.passed_all:
            return f"<p><strong><pre style='display: inline;'>{self.name}</pre></strong> passed!</p>"
        else:
            ret = f"<p><strong style='color: red;'><pre style='display: inline;'>{self.name}</pre> results:</strong></p>"
            for tcr in self.test_case_results:
                ret += f"<p><strong><pre style='display: inline;'>{tcr.test_case.name}</pre> result:</strong></p>"
                ret += f"<pre>{indent(tcr.message, '    ')}</pre>"
                if tcr.passed and tcr.test_case.success_message is not None:
                    ret += f"<p><strong><pre style='display: inline;'>{tcr.test_case.name}</pre> message:</strong> {tcr.test_case.success_message}</p>"
                if not tcr.passed and tcr.test_case.failure_message is not None:
                    ret += f"<p><strong><pre style='display: inline;'>{tcr.test_case.name}</pre> message:</strong> {tcr.test_case.failure_message}</p>"
            return ret

    def __repr__(self):
        return self.summary()

    # @abstractmethod
    def __init__(self, name, path, test_cases, all_or_nothing=True):
        self.name = name
        self.path = path
        self.test_cases = test_cases
        self.all_or_nothing = all_or_nothing
        self.test_case_results = []
        self._score = None

    @staticmethod
    def resolve_test_file_points(total_points, test_cases):
        if isinstance(total_points, list):
            if len(total_points) != len(test_cases):
                raise ValueError("Points specified in test has different length than number of test cases")
            test_cases = [tc._replace(points=pt) for tc, pt in zip(test_cases, total_points)]
            total_points = None
        elif total_points is not None and not isinstance(total_points, (int, float)):
            raise TypeError(f"Test spec points has invalid type: {total_points}")

        point_values = []
        for i, test_case in enumerate(test_cases):
            if test_case.points is not None:
                assert type(test_case.points) in (int, float), f"Invalid point type: {type(test_case.points)}"
                point_values.append(test_case.points)
            else:
                point_values.append(None)

        pre_specified = sum(p for p in point_values if p is not None)
        if total_points is not None:
            if pre_specified > total_points:
                raise ValueError(f"More points specified in test cases that allowed for test")
            else:
                try:
                    per_remaining = (total_points - pre_specified) / sum(1 for p in point_values if p is None)
                except ZeroDivisionError:
                    per_remaining = 0.0
        else:
            if pre_specified == 0 and all(p in (0, None) for p in point_values):
                # if only zeros specified, assume test worth 1 pt and divide amongst nonzero cases
                try:
                    per_remaining = 1 / sum(p is None for p in point_values)
                except ZeroDivisionError:
                    per_remaining = 0.0
            elif pre_specified == 0:
                per_remaining = 1 / len(point_values)
            else:
                # assume all other tests are worth 0 points
                per_remaining = 0.0

        point_values = [p if p is not None else per_remaining for p in point_values]
        return [tc._replace(points=p) for tc, p in zip(test_cases, point_values)]

    @property
    def passed_all(self):
        return all(tcr.passed for tcr in self.test_case_results)

    @property
    def passed_all_public(self):
        return all(tcr.passed for tcr in self.test_case_results if not tcr.test_case.hidden)

    @property
    def all_public(self):
        return all(not tc.hidden for tc in self.test_cases)

    @property
    def grade(self):
        if self.all_or_nothing and not self.passed_all:
            return 0
        elif self.all_or_nothing and self.passed_all:
            return 1
        else:
            return sum(tcr.test_case.points for tcr in self.test_case_results if tcr.passed) / \
                sum(tc.points for tc in self.test_cases)

    @property
    def score(self):
        if self._score is not None:
            return self._score
        return sum(tcr.test_case.points for tcr in self.test_case_results if tcr.passed)

    @property
    def possible(self):
        return sum(tc.points for tc in self.test_cases)

    def update_score(self, new_score):
        self._score = new_score

    def to_dict(self):
        return {
            "score": self.score,
            "possible": self.possible,
            "name": self.name,
            "path": self.path,
            "test_cases": [tc._asdict() for tc in self.test_cases],
            "all_or_nothing": self.all_or_nothing,
            "test_case_results": [tcr._asdict() for tcr in self.test_case_results],
        }

    def summary(self, public_only=False):
        if (not public_only and self.passed_all) or (public_only and self.passed_all_public):
            return f"{self.name} results: All test cases passed!"

        tcrs = self.test_case_results
        if public_only:
            tcrs = [tcr for tcr in tcrs if not tcr.test_case.hidden]
        
        tcr_summaries = []
        for tcr in tcrs:
            smry = f"{tcr.test_case.name} result:\n"
            smry += f"{indent(tcr.message.strip(), '    ')}\n\n"
            if tcr.passed and tcr.test_case.success_message is not None:
                smry += f"{tcr.test_case.name} message: {tcr.test_case.success_message}\n\n"
            if not tcr.passed and tcr.test_case.failure_message is not None:
                smry += f"{tcr.test_case.name} message: {tcr.test_case.failure_message}\n\n"

            tcr_summaries.append(smry.strip())

        return f"{self.name} results:\n" + indent("\n\n".join(tcr_summaries), "    ")

    @classmethod
    @abstractmethod
    def from_file(cls, path):
        ...

    @abstractmethod
    def run(self, global_environment):
        ...



"""
Otter-Grader version and printable logo
"""

import sys

from textwrap import dedent

__version__ = "3.1.2"

LOGO_WITH_VERSION = f"""\
  _________        __          __               
 /  _____  \    __|  |__    __|  |__               
|  /     \  |  |__    __|  |__    __|   _______    _  _____
| |       | |     |  |        |  |     |  ___  |  | |/ ____|
| |       | |     |  |        |  |     | |___| |  |   /    
| |       | |     |  |        |  |     | ______|  |  |
|  \_____/  |     |  |_       |  |_    | |_____   |  |
 \_________/       \ __|       \ __|    \______|  |__|
                                                v{__version__}
"""

def print_version_info(logo=False):
    """
    Prints the Otter logo and version information to stdout

    Args:
        logo (``bool``, optional): whether to print the logo
    """
    if logo:
        print(LOGO_WITH_VERSION)
    print(dedent(f"""\
        Python version: {".".join(str(v) for v in sys.version_info[:3])}
        Otter-Grader version: {__version__}"""))



"""
Grading plugins for Otter
"""

import importlib

from .abstract_plugin import AbstractOtterPlugin, PluginEventNotSupportedException
from ..utils import print_full_width


class PluginCollection:
    """
    Class for loading, organizing, and running plugins during grading. This class is instantiated with
    a list of plugin names, which should be importable strings that evaluate to objects that inherit 
    from ``otter.plugins.AbstractOtterPlugin``.

    When this class is instantiated, each plugin is imported and passed its configurations specified
    in the ``otter_config.json``. Plugins should be listed in ``otter_config.json`` in the ``plugins``
    key:

    .. code-block:: json

        {
            "plugins": [
                "some_otter_plugin_package.SomeOtterPlugin",
                {
                    "some_other_otter_plugin_package.SomeOtherOtterPlugin": {
                        "some_key": "some_value"
                    }
                }
            ]
        }

    Args:
        plugin_names (``list[Union[str,dict[str:Any]]]``): the importable names of plugin classes (e.g. 
            ``some_package.SomePlugin``) and their configurations
        submission_path (``str``): the absolute path to the submission being graded
        submission_metadata (``dict[str:Any]``): submission metadata
    """

    @staticmethod
    def _parse_plugin_config(plugin_config):
        if not isinstance(plugin_config, list):
            raise ValueError(f"Invalid plugin config: {plugin_config}")
        
        result = []
        for plg in plugin_config:
            if isinstance(plg, str):
                result.append({
                    "plugin": plg,
                    "config": {},
                })
            elif isinstance(plg, dict):
                keys = list(plg.keys())
                if not len(keys) == 1:
                    raise ValueError(f"Invalid plugin specification: {plg}")
                result.append({
                    "plugin": keys[0],
                    "config": plg[keys[0]],
                })

        return result

    def __init__(self, plugins, submission_path, submission_metadata):
        self._plugin_config = self._parse_plugin_config(plugins)
        self._plugins = None

        self._subm_path = submission_path
        self._subm_meta = submission_metadata

        self._plugins = self._load_plugins(self._plugin_config, submission_path, submission_metadata)

    @property
    def _plugin_names(self):
        """
        The importable names of all of the plugins tracked
        """
        return [p["plugin"] for p in self._plugin_config]

    @staticmethod
    def _load_plugins(plugin_config, submission_path, submission_metadata):
        """
        Loads each plugin in ``self._plugin_config`` by importing it with ``importlib`` and creating
        and instance with the ``submission_metadata`` and the configurations from ``self._plugin_config``
        for that plugin. Sets ``self._plugins`` to be the list of imported and instantiated plugins.

        Args:
            submission_path (``str``): the absolute path to the submission being graded
            submission_metadata (``dict``): submission metadata

        Returns:
            ``list[AbstractOtterPlugin]``: the list of instantiated plugins
        """
        plugins = []
        for plg_cfg in plugin_config:
            plg, cfg = plg_cfg["plugin"], plg_cfg["config"]
            module, class_ = ".".join(plg.split(".")[:-1]), plg.split(".")[-1]
            module = importlib.import_module(module)
            class_ = getattr(module, class_)

            # get the config key for the plugin
            # plugin_cfg = plugin_config.get(class_.PLUGIN_CONFIG_KEY, {})
            plugin = class_(submission_path, submission_metadata, cfg)
            plugins.append(plugin)

        return plugins
    
    def add_new_plugins(self, raw_plugin_config):
        """
        Add any new plugins specified in ``raw_plugin_config`` to this plugin collection. Any plugins
        listed that have already been isntatiated here are *not* added.

        Args:
            raw_plugin_config (``list[Union[str,dict[str:Any]]]``): the importable names of plugin classes (e.g. 
            ``some_package.SomePlugin``) and their configurations
        """
        plg_cfg = self._parse_plugin_config(raw_plugin_config)
        for i, plg in list(enumerate(plg_cfg))[::-1]:
            name = plg["plugin"]
            if any(c["plugin"] == plg["plugin"] for c in self._plugin_config):
                plg_cfg.pop(i)
        
        self._plugin_config.extend(plg_cfg)
        self._plugins.extend(self._load_plugins(plg_cfg, self._subm_path, self._subm_meta))

    def run(self, event, *args, **kwargs):
        """
        Runs the method ``event`` of each plugin in this collection. Passes ``args`` and ``kwargs``
        to this method. Ignores plugins that raise ``PluginEventNotSupportedException`` for this event.

        Args:
            event (``str``): name of the method of the plugin to run
            *args, **kwargs (any): arguments for the method
        
        Returns:
            ``list[Any]``: the values returned by each plugin for the called event
        """
        # TODO: logging to stdout
        rets = []
        for plugin in self._plugins:
            try:
                if hasattr(plugin, event):
                    ret = getattr(plugin, event)(*args, **kwargs)
                    rets.append(ret)
                else:
                    rets.append(None)
            except PluginEventNotSupportedException:
                rets.append(None)
        return rets

    def before_execution(self, submission):
        """
        Runs the ``before_execution`` event for each plugin, composing the results of each (i.e. the
        transformed notebook returned by one plugin is passed to the next plugin).

        Args:
            submission (``Union[str,nbformat.NotebookNode]``): the submission to be executed
        """
        event = "before_execution"
        for plugin in self._plugins:
            try:
                if hasattr(plugin, event):
                    submission = getattr(plugin, event)(submission)
            except PluginEventNotSupportedException:
                pass
        return submission

    def generate_report(self):
        """
        Runs the ``generate_report`` event of each plugin, formatting and concatenating them into a
        single string and returning it.

        Returns:
            ``str``: the plugin report
        """
        reports = self.run("generate_report")
        if not any(isinstance(r, str) for r in reports):
            return ""

        # header = "=" * 35 + " PLUGIN REPORT " + "=" * 35
        header = print_full_width("=", mid_text="PLUGIN REPORT", ret_str=True)
        footer = "=" * len(header)

        report = header
        for r, plg in zip(reports, self._plugin_names):
            if not isinstance(r, str):
                continue
            # title = f" {plg} Report "
            # dashes = len(header) - len(title)
            # if dashes > 4:
            #     if dashes % 2 == 0:
            #         ld, rd = dashes // 2, dashes // 2
            #     else:
            #         ld, rd = dashes // 2, dashes // 2 + 1
            # else:
            #     ld, rd = 2, 2

            # title = "-" * ld + title + "-" * rd
            title = print_full_width("-", mid_text=f"{plg} Report", ret_str=True)

            body = "\n" + title + "\n" + r + "\n"

            report += "\n" + body
        
        report += "\n" + footer

        return report



"""
ottr test adapters for Otter Assign
"""

import re
import pprint
import yaml
import nbformat

from collections import namedtuple

from ..constants import BEGIN_TEST_CONFIG_REGEX, END_TEST_CONFIG_REGEX, OTTR_TEST_FILE_TEMPLATE, \
    OTTR_TEST_NAME_REGEX, TEST_REGEX
from ..tests import write_test
from ..utils import get_source, lock

from ...test_files.abstract_test import TestFile


Test = namedtuple('Test', ['name', 'hidden', 'points', 'body', 'success_message', 'failure_message'])


def read_test(cell, question, assignment, rmd=False):
    """
    Returns the contents of a test as a ``(name, hidden, body)`` named tuple
    
    Args:
        cell (``nbformat.NotebookNode``): a test cell
        question (``dict``): question metadata
        assignment (``otter.assign.assignment.Assignment``): the assignment configurations
        rmd (``bool``, optional): whether the cell is from an Rmd file; if true, the first and last
            lines of ``cell``'s source are trimmed, since they should be backtick delimeters

    Returns:
        ``Test``: test named tuple
    """
    if rmd:
        source = get_source(cell)[1:-1]
    else:
        source = get_source(cell)

    if source[0].lstrip().startswith("#"):
        hidden = bool(re.search(r"hidden", source[0], flags=re.IGNORECASE))
    else:
        hidden = False

    i = 0 if hidden else -1
    if rmd:
        i = 0

    if re.match(BEGIN_TEST_CONFIG_REGEX, source[0], flags=re.IGNORECASE):
        for i, line in enumerate(source):
            if re.match(END_TEST_CONFIG_REGEX, line, flags=re.IGNORECASE):
                break
        config = yaml.full_load("\n".join(source[1:i]))
        assert isinstance(config, dict), f"Invalid test config in cell {cell}"
    else:
        config = {}

    test_name = config.get("name", None)
    hidden = config.get("hidden", hidden)
    points = config.get("points", None)
    success_message = config.get("success_message", None)
    failure_message = config.get("failure_message", None)

    return Test(test_name, hidden, points, '\n'.join(source[i+1:]), success_message, failure_message)


def gen_test_cell(question, tests, tests_dict, assignment):
    """
    Parses a list of test named tuples and creates a single test file. Adds this test file as a value
    to ``tests_dict`` with a key corresponding to the test's name, taken from ``question``. Returns
    a code cell that runs the check on this test.
    
    Args:
        question (``dict``): question metadata
        tests (``list`` of ``Test``): tests to be written
        tests_dict (``dict``): the tests for this assignment
        assignment (``otter.assign.assignment.Assignment``): the assignment configurations

    Returns:
        ``nbformat.NotebookNode``: code cell calling ``ottr::check`` on this test
    """
    cell = nbformat.v4.new_code_cell()
    cell.source = ['. = ottr::check("tests/{}.R")'.format(question['name'])]

    points = question.get('points', None)
    if isinstance(points, dict):
        points = points.get('each', 1) * len(tests)
    elif isinstance(points, list):
        if len(points) != len(tests):
            raise ValueError(
                f"Error in question {question['name']}: length of 'points' is {len(points)} but there "
                f"are {len(tests)} tests")

    # check for errors in resolving points
    tests = TestFile.resolve_test_file_points(points, tests)

    test = gen_suite(question['name'], tests, points)

    tests_dict[question['name']] = test
    lock(cell)
    return cell


def gen_suite(name, tests, points):
    """
    Generates an R-formatted test file for ottr

    Args:
        name (``str``): the test name
        tests (``list`` of ``Test``): the test case named tuples that define this test file
        points (``float`` or ``int`` or ``list`` of ``float`` or ``int``): th points per question

    Returns:
        ``str``: the rendered R test file
    """
    template_data = {'name': name, 'test_cases': tests}
    return OTTR_TEST_FILE_TEMPLATE.render(**template_data)


def remove_hidden_tests_from_dir(nb, test_dir, assignment, use_files=True):
    """
    Rewrites test files in a directory to remove hidden tests
    
    Args:
        nb (``nbformat.NotebookNode``): the student notebook
        test_dir (``pathlib.Path``): path to test files directory
        assignment (``otter.assign.assignment.Assignment``): the assignment configurations
        use_files (``bool``, optional): ignored for R assignments
    """
    for f in test_dir.iterdir():
        if f.suffix != '.R':
            continue

        with open(f) as f2:
            test = f2.read()
        
        test = re.sub(r"    ottr::TestCase\$new\(\s*hidden = TRUE[\w\W]+?^    \),?", "", test, flags=re.MULTILINE)
        test = re.sub(r",(\s*  \))", r"\1", test, flags=re.MULTILINE)  # removes a trailing comma if present

        write_test({}, f, test, use_file=True)
